"""
    Copyright 2017 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import asyncio
import inspect
import os
import subprocess
import warnings
from collections import abc
from functools import reduce
from typing import TYPE_CHECKING, Any, Callable, Dict, FrozenSet, List, Optional, Tuple, Type, TypeVar

import inmanta.ast.type as inmanta_type
from inmanta import const, protocol, util
from inmanta.ast import (
    CompilerException,
    LocatableString,
    Location,
    Namespace,
    Range,
    RuntimeException,
    TypeNotFoundException,
    WithComment,
)
from inmanta.ast.type import NamedType
from inmanta.config import Config
from inmanta.execute.proxy import DynamicProxy
from inmanta.execute.runtime import QueueScheduler, Resolver, ResultVariable
from inmanta.execute.util import Unknown
from inmanta.stable_api import stable_api
from inmanta.warnings import InmantaWarning

T = TypeVar("T")

if TYPE_CHECKING:
    from inmanta.ast.statements import DynamicStatement
    from inmanta.ast.statements.call import FunctionCall
    from inmanta.compiler import Compiler


class PluginDeprecationWarning(InmantaWarning):
    pass


@stable_api
class Context(object):
    """
    An instance of this class is used to pass context to the plugin
    """

    __client: Optional["protocol.Client"] = None
    __sync_client = None

    @classmethod
    def __get_client(cls) -> "protocol.Client":
        if cls.__client is None:
            cls.__client = protocol.Client("compiler")
        return cls.__client

    def __init__(
        self, resolver: Resolver, queue: QueueScheduler, owner: "FunctionCall", plugin: "Plugin", result: ResultVariable
    ) -> None:
        self.resolver = resolver
        self.queue = queue
        self.owner = owner
        self.plugin = plugin
        self.result = result
        self.compiler = queue.get_compiler()

    def get_resolver(self) -> Resolver:
        return self.resolver

    def get_type(self, name: LocatableString) -> inmanta_type.Type:
        """
        Get a type from the configuration model.
        """
        try:
            return self.queue.get_types()[str(name)]
        except KeyError:
            raise TypeNotFoundException(name, self.owner.namespace)

    def get_queue_scheduler(self) -> QueueScheduler:
        return self.queue

    def get_environment_id(self) -> str:
        env = str(Config.get("config", "environment", None))

        if env is None:
            raise Exception("The environment of the model should be configured in config>environment")

        return env

    def get_compiler(self) -> "Compiler":
        return self.queue.get_compiler()

    def get_data_dir(self) -> str:
        """
        Get the path to the data dir (and create if it does not exist yet
        """
        data_dir = os.path.join("data", self.plugin.namespace.get_full_name())

        if not os.path.exists(data_dir):
            os.makedirs(data_dir, exist_ok=True)

        return data_dir

    def get_client(self) -> "protocol.Client":
        return self.__class__.__get_client()

    def get_sync_client(self) -> "protocol.SyncClient":
        if self.__class__.__sync_client is None:
            self.__class__.__sync_client = protocol.SyncClient("compiler")
        return self.__class__.__sync_client

    def run_sync(self, function: Callable[[], abc.Awaitable[T]], timeout: int = 5) -> T:
        """
        Execute the async function and return its result. This method uses this thread's current (not running) event loop if
        there is one, otherwise it creates a new one. The main use for this function is to use the inmanta internal rpc to
        communicate with the server.

        :param function: The async function to execute. This function should return a yieldable object.
        :param timeout: A timeout for the async function.
        :return: The result of the async call.
        :raises ConnectionRefusedError: When the function timeouts this exception is raised.
        """
        with_timeout: abc.Awaitable[T] = asyncio.wait_for(function(), timeout)
        try:
            return util.ensure_event_loop().run_until_complete(with_timeout)
        except TimeoutError:
            raise ConnectionRefusedError()


@stable_api
class PluginMeta(type):
    """
    A metaclass that keeps track of concrete plugin subclasses. This class is responsible for all plugin registration.
    """

    def __new__(cls, name: str, bases: Tuple[type, ...], dct: Dict[str, object]) -> Type:
        subclass = type.__new__(cls, name, bases, dct)
        if hasattr(subclass, "__function_name__"):
            cls.add_function(subclass)
        return subclass

    __functions: Dict[str, Type["Plugin"]] = {}

    @classmethod
    def add_function(cls, plugin_class: Type["Plugin"]) -> None:
        """
        Add a function plugin class
        """
        cls.__functions[plugin_class.__fq_plugin_name__] = plugin_class

    @classmethod
    def get_functions(cls) -> Dict[str, "Type[Plugin]"]:
        """
        Get all functions that are registered
        """
        return dict(cls.__functions)

    @classmethod
    def clear(cls, inmanta_module: Optional[str] = None) -> None:
        """
        Clears registered plugin functions.

        :param inmanta_module: Clear plugin functions for a specific inmanta module. If omitted, clears all registered plugin
            functions.
        """
        if inmanta_module is not None:
            top_level: str = f"{const.PLUGINS_PACKAGE}.{inmanta_module}"
            cls.__functions = {
                fq_name: plugin_class
                for fq_name, plugin_class in cls.__functions.items()
                if plugin_class.__module__ != top_level and not plugin_class.__module__.startswith(f"{top_level}.")
            }
        else:
            cls.__functions = {}


class PluginArgument:
    """
    Represents the argument of an Inmanta plugin.
    """

    # Marker used to indicate that a plugin argument has no default value.
    NO_DEFAULT_VALUE_SET = object()

    def __init__(
        self, arg_name: str, arg_type: object, is_kw_only_argument: bool, default_value: object = NO_DEFAULT_VALUE_SET
    ) -> None:
        self.arg_name = arg_name
        self.arg_type = arg_type
        self.is_kw_only_argument = is_kw_only_argument
        self._default_value = default_value

    @property
    def default_value(self) -> Optional[object]:
        if not self.has_default_value():
            raise Exception("PluginArgument doesn't have a default value")
        return self._default_value

    def has_default_value(self) -> bool:
        """
        Return True iff this plugin argument has a default value set.
        """
        return self._default_value is not self.NO_DEFAULT_VALUE_SET


class Plugin(NamedType, WithComment, metaclass=PluginMeta):
    """
    This class models a plugin that can be called from the language.
    """

    deprecated: bool = False
    replaced_by: Optional[str] = None

    def __init__(self, namespace: Namespace) -> None:
        self.ns = namespace
        self.namespace = namespace

        # The index of the Context attribute.
        self._context: int = -1
        self._return = None

        self.arguments: List[PluginArgument]
        if hasattr(self.__class__, "__function__"):
            self.arguments = self._load_signature(self.__class__.__function__)
        else:
            self.arguments = []

        self.new_statement = None

        filename: Optional[str] = inspect.getsourcefile(self.__class__.__function__)
        assert filename is not None
        try:
            line: int = inspect.getsourcelines(self.__class__.__function__)[1] + 1
        except OSError:
            # In case of bytecompiled code there is no source line
            line = 1

        if self.__class__.__function__.__doc__:
            self.comment = self.__class__.__function__.__doc__

        self.location = Location(filename, line)

    def normalize(self) -> None:
        self.resolver = self.namespace
        self.argtypes = [self.to_type(arg.arg_type, self.namespace) for arg in self.arguments]
        self.returntype = self.to_type(self._return, self.namespace)

    def _load_signature(self, function: Callable[..., object]) -> List[PluginArgument]:
        """
        Load the signature from the given python function
        """
        arg_spec = inspect.getfullargspec(function)

        # Calculate at which index the default values start (for the non-keyword-only attributes)
        default_start_for_args: Optional[int]
        if arg_spec.defaults is not None:
            default_start_for_args = len(arg_spec.args) - len(arg_spec.defaults)
        else:
            default_start_for_args = None

        arguments: List[PluginArgument] = []

        def process_kw_only_args(argument_name: str, annotation: object) -> PluginArgument:
            if arg_spec.kwonlydefaults and argument_name in arg_spec.kwonlydefaults:
                default_value = arg_spec.kwonlydefaults[argument_name]
                return PluginArgument(argument_name, annotation, is_kw_only_argument=True, default_value=default_value)
            else:
                return PluginArgument(argument_name, annotation, is_kw_only_argument=True)

        def process_regular_args(index: int, argument_name: str, annotation: object) -> PluginArgument:
            if default_start_for_args is not None and default_start_for_args <= index:
                default_value = arg_spec.defaults[default_start_for_args - index]
                return PluginArgument(argument_name, annotation, is_kw_only_argument=False, default_value=default_value)
            else:
                return PluginArgument(argument_name, annotation, is_kw_only_argument=False)

        def process_arg(index: int, argument_name: str, is_kwonly_arg: bool) -> None:
            if argument_name not in arg_spec.annotations:
                raise CompilerException(f"All arguments of plugin '{function.__name__}' should be annotated")
            annotation = arg_spec.annotations[argument_name]
            if annotation == Context:
                self._context = index
            else:
                if is_kwonly_arg:
                    plugin_argument = process_kw_only_args(argument_name, annotation)
                else:
                    plugin_argument = process_regular_args(index, argument_name, annotation)
                arguments.append(plugin_argument)

        # Process regular arguments
        for i in range(len(arg_spec.args)):
            process_arg(i, arg_spec.args[i], is_kwonly_arg=False)
        # Process keyword-only arguments
        for i in range(len(arg_spec.kwonlyargs)):
            process_arg(i, arg_spec.kwonlyargs[i], is_kwonly_arg=True)

        if "return" in arg_spec.annotations:
            self._return = arg_spec.annotations["return"]

        return arguments

    def get_signature(self) -> str:
        """
        Generate the signature of this plugin
        """
        arg_list = []
        for arg in self.arguments:
            if arg.has_default_value():
                arg_list.append("%s: %s=%s" % (arg.arg_name, arg.arg_type, str(arg.default_value)))
            else:
                arg_list.append("%s: %s" % (arg.arg_name, arg.arg_type))

        args = ", ".join(arg_list)

        if self._return is None:
            return "%s(%s)" % (self.__class__.__function_name__, args)
        return "%s(%s) -> %s" % (self.__class__.__function_name__, args, self._return)

    def to_type(self, arg_type: Optional[object], resolver: Namespace) -> Optional[inmanta_type.Type]:
        """
        Convert a string representation of a type to a type
        """
        if arg_type is None:
            return None

        if not isinstance(arg_type, str):
            raise CompilerException(
                "bad annotation in plugin %s::%s, expected str but got %s (%s)"
                % (self.ns, self.__class__.__function_name__, type(arg_type), arg_type)
            )

        if arg_type == "any":
            return None

        if arg_type == "expression":
            return None

        # quickfix issue #1774
        allowed_element_type: inmanta_type.Type = inmanta_type.Type()
        if arg_type == "list":
            return inmanta_type.TypedList(allowed_element_type)
        if arg_type == "dict":
            return inmanta_type.TypedDict(allowed_element_type)

        plugin_line: Range = Range(self.location.file, self.location.lnr, 1, self.location.lnr + 1, 1)
        locatable_type: LocatableString = LocatableString(arg_type, plugin_line, 0, None)

        # stack of transformations to be applied to the base inmanta_type.Type
        # transformations will be applied right to left
        transformation_stack: List[Callable[[inmanta_type.Type], inmanta_type.Type]] = []

        if locatable_type.value.endswith("?"):
            locatable_type.value = locatable_type.value[0:-1]
            transformation_stack.append(inmanta_type.NullableType)

        if locatable_type.value.endswith("[]"):
            locatable_type.value = locatable_type.value[0:-2]
            transformation_stack.append(inmanta_type.TypedList)

        return reduce(lambda acc, transform: transform(acc), reversed(transformation_stack), resolver.get_type(locatable_type))

    def _is_instance(self, value: object, arg_type: Type[object]) -> bool:
        """
        Check if value is of arg_type
        """
        if arg_type is None:
            return True

        if hasattr(arg_type, "validate"):
            return arg_type.validate(value)

        return isinstance(value, arg_type)

    def check_args(self, args: List[object], kwargs: Dict[str, object]) -> bool:
        """
        Check if the arguments of the call match the function signature
        """
        max_arg = len(self.arguments)
        if len(args) + len(kwargs) > max_arg:
            raise Exception(
                "Incorrect number of arguments for %s. Expected at most %d, got %d"
                % (self.get_signature(), max_arg, len(args) + len(kwargs))
            )
        # check for missing arguments
        required_args: FrozenSet[str] = frozenset(arg.arg_name for arg in self.arguments if not arg.has_default_value())
        present_positional_args: FrozenSet[str] = frozenset(arg.arg_name for arg in self.arguments[: len(args)])
        present_kwargs: FrozenSet[str] = frozenset(kwargs.keys())
        missing_args = required_args.difference({*present_positional_args, *present_kwargs})
        if missing_args:
            raise RuntimeException(
                None,
                "Missing %d required arguments for %s(): %s"
                % (len(missing_args), self.__class__.__function_name__, ",".join(missing_args)),
            )
        # check for kwargs overlap with positional arguments
        overlapping_args: FrozenSet[str] = present_kwargs.intersection(present_positional_args)
        if overlapping_args:
            raise RuntimeException(
                None,
                "Multiple values for %s in %s()" % (",".join(overlapping_args), self.__class__.__function_name__),
            )

        def is_valid(expected_arg: PluginArgument, expected_type: Type[object], arg: object) -> bool:
            if isinstance(arg, Unknown):
                return False

            if not self._is_instance(arg, expected_type):
                raise Exception(
                    "Invalid type for argument %s of '%s', it should be %s and %s given."
                    % (expected_arg.arg_name, self.__class__.__function_name__, expected_arg.arg_type, arg.__class__.__name__)
                )
            return True

        for i in range(len(args)):
            if not is_valid(self.arguments[i], self.argtypes[i], args[i]):
                return False
        arg_types: Dict[str, Tuple[PluginArgument, Optional[inmanta_type.Type]]] = {
            arg.arg_name: (arg, self.argtypes[i]) for i, arg in enumerate(self.arguments)
        }
        for k, v in kwargs.items():
            try:
                (expected_arg, expected_type) = arg_types[k]
                if not is_valid(expected_arg, expected_type, v):
                    return False
            except KeyError:
                raise RuntimeException(None, "Invalid keyword argument '%s' for '%s()'" % (k, self.__class__.__function_name__))
        return True

    def emit_statement(self) -> "DynamicStatement":
        """
        This method is called to determine if the plugin call pushes a new
        statement
        """
        return self.new_statement

    def is_accept_unknowns(self) -> bool:
        return self.opts["allow_unknown"]

    def check_requirements(self) -> None:
        """
        Check if the plug-in has all it requires
        """
        if "bin" in self.opts and self.opts["bin"] is not None:
            for _bin in self.opts["bin"]:
                p = subprocess.Popen(["bash", "-c", "type -p %s" % _bin], stdout=subprocess.PIPE)
                result = p.communicate()

                if len(result[0]) == 0:
                    raise Exception("%s requires %s to be available in $PATH" % (self.__function_name__, _bin))

    @classmethod
    def deprecate_function(cls, replaced_by: Optional[str] = None) -> None:
        cls.deprecated = True
        cls.replaced_by = replaced_by

    def __call__(self, *args: object, **kwargs: object) -> object:
        """
        The function call itself
        """
        if self.deprecated:
            msg: str = f"Plugin '{self.__function_name__}' in module '{self.__module__}' is deprecated."
            if self.replaced_by:
                msg += f" It should be replaced by '{self.replaced_by}'."
            warnings.warn(PluginDeprecationWarning(msg))
        self.check_requirements()

        def new_arg(arg: object) -> object:
            if isinstance(arg, Context):
                return arg
            elif isinstance(arg, Unknown) and self.is_accept_unknowns():
                return arg
            else:
                return DynamicProxy.return_value(arg)

        new_args = [new_arg(arg) for arg in args]
        new_kwargs = {k: new_arg(v) for k, v in kwargs.items()}

        value = self.call(*new_args, **new_kwargs)

        value = DynamicProxy.unwrap(value)

        if self.returntype is not None and not isinstance(value, Unknown):
            valid = False
            exception = None

            try:
                valid = value is None or self._is_instance(value, self.returntype)
            except RuntimeException as e:
                raise e
            except Exception as exp:
                exception = exp

            if not valid:
                msg = ""
                if exception is not None:
                    msg = "\n\tException details: " + str(exception)

                raise Exception(
                    "Plugin %s should return value of type %s ('%s' was returned) %s"
                    % (self.__class__.__function_name__, self.returntype, value, msg)
                )

        return value

    def get_full_name(self) -> str:
        return "%s::%s" % (self.ns.get_full_name(), self.__class__.__function_name__)

    def type_string(self) -> str:
        return self.get_full_name()


@stable_api
class PluginException(Exception):
    """
    Base class for custom exceptions raised from a plugin.
    """

    def __init__(self, message: str) -> None:
        self.message = message


@stable_api
def plugin(
    function: Optional[Callable] = None,
    commands: Optional[List[str]] = None,
    emits_statements: bool = False,
    allow_unknown: bool = False,
) -> Callable:
    """
    Python decorator to register functions with inmanta as plugin

    :param function: The function to register with inmanta. This is the first argument when it is used as decorator.
    :param commands: A list of command paths that need to be available. Inmanta raises an exception when the command is
                     not available.
    :param emits_statements: Set to true if this plugin emits new statements that the compiler should execute. This is only
                             required for complex plugins such as integrating a template engine.
    :param allow_unknown: Set to true if this plugin accepts Unknown values as valid input.
    """

    def curry_name(
        name: Optional[str] = None,
        commands: Optional[List[str]] = None,
        emits_statements: bool = False,
        allow_unknown: bool = False,
    ) -> Callable:
        """
        Function to curry the name of the function
        """

        def call(fnc):
            """
            Create class to register the function and return the function itself
            """

            def wrapper(self, *args: object, **kwargs: object) -> Any:
                """
                Python will bind the function as method into the class
                """
                return fnc(*args, **kwargs)

            nonlocal name, commands, emits_statements

            if name is None:
                name = fnc.__name__

            ns_parts = str(fnc.__module__).split(".")
            ns_parts.append(name)
            if ns_parts[0] != const.PLUGINS_PACKAGE:
                raise Exception("All plugin modules should be loaded in the %s package" % const.PLUGINS_PACKAGE)

            fq_plugin_name = "::".join(ns_parts[1:])

            dictionary = {}
            dictionary["__module__"] = fnc.__module__

            dictionary["__function_name__"] = name
            dictionary["__fq_plugin_name__"] = fq_plugin_name

            dictionary["opts"] = {"bin": commands, "emits_statements": emits_statements, "allow_unknown": allow_unknown}
            dictionary["call"] = wrapper
            dictionary["__function__"] = fnc

            bases = (Plugin,)
            fnc.__plugin__ = PluginMeta.__new__(PluginMeta, name, bases, dictionary)
            return fnc

        return call

    if function is None:
        return curry_name(commands=commands, emits_statements=emits_statements, allow_unknown=allow_unknown)

    elif isinstance(function, str):
        return curry_name(function, commands=commands, emits_statements=emits_statements, allow_unknown=allow_unknown)

    elif function is not None:
        fnc = curry_name(commands=commands, emits_statements=emits_statements, allow_unknown=allow_unknown)
        return fnc(function)


@stable_api
def deprecated(
    function: Optional[Callable] = None, *, replaced_by: Optional[str] = None, **kwargs: abc.Mapping[str, object]
) -> Callable:
    """
    the kwargs are currently ignored but where added in case we want to add something later on.
    """

    def inner(fnc: Callable):
        if hasattr(fnc, "__plugin__"):
            fnc.__plugin__.deprecate_function(replaced_by)
        else:
            raise Exception(
                f"Can not deprecate '{fnc.__name__}': The '@deprecated' decorator should be used in combination with the "
                f"'@plugin' decorator and should be placed at the top."
            )
        return fnc

    if function is not None:
        return inner(function)
    return inner



"""
    Copyright 2020 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
from asyncpg import Connection

DISABLED = False


async def update(connection: Connection) -> None:
    await connection.execute(
        """
        -- Ensure agent records don't get deleted when the referenced agent_instance is deleted
        ALTER TABLE public.agent DROP CONSTRAINT agent_id_primary_fkey;
        ALTER TABLE  public.agent
        ADD CONSTRAINT agent_id_primary_fkey
        FOREIGN KEY (id_primary)
        REFERENCES public.agentinstance(id)
        ON DELETE RESTRICT;

        -- Used by data.Agent.expire_all()
        CREATE INDEX agent_id_primary_index ON agent (id_primary) WHERE (id_primary IS NULL);
        -- Used by data.AgentProcess.expire_all()
        CREATE INDEX agentprocess_expired_index ON agentprocess (expired) WHERE (expired IS NULL);
        -- Used by data.AgentProcess.cleanup()
        CREATE INDEX agentprocess_env_hostname_expired_index ON agentprocess (environment, hostname, expired);
        -- Used by data.AgentInstance.expire_all()
        CREATE INDEX agentinstance_expired_index ON agentinstance (expired) WHERE (expired IS NULL);
        """
    )



"""
    Copyright 2018 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import logging
import os
import shutil
import subprocess
import tempfile
from typing import Optional

PG_CTL_BIN = "pg_ctl"
INITDB_BIN = "initdb"


LOGGER = logging.getLogger(__name__)


def find_executable(executable: str) -> Optional[str]:
    """
    Scan PATH for an executable.
    """
    for path in os.environ.get("PATH", "").split(os.pathsep):
        path = os.path.abspath(path)
        executable_path = os.path.join(path, executable)
        if os.path.isfile(executable_path):
            return executable_path

    # fall back to pg_config
    try:
        bindir = subprocess.check_output(["pg_config", "--bindir"], universal_newlines=True).strip()
        binpath = os.path.join(bindir, executable)
        if os.path.isfile(binpath):
            return binpath
    except FileNotFoundError:
        return None

    return None


class PostgresProc(object):
    def __init__(
        self, port: int, pg_ctl_bin: Optional[str] = None, initdb_bin: Optional[str] = None, db_path: Optional[str] = None
    ) -> None:
        self.port = port
        self.db_path = db_path
        if self.db_path:
            if os.path.exists(self.db_path) and os.path.isfile(self.db_path):
                raise AssertionError("DB path should be a directory, but it is a file.")

        ctl_bin = pg_ctl_bin or find_executable(PG_CTL_BIN)
        assert ctl_bin, f"Could not find '{PG_CTL_BIN}' or pg_config in system PATH. Make sure you have PostgreSQL installed."
        self.pg_ctl_bin: str = ctl_bin

        initdb_bin = initdb_bin or find_executable(INITDB_BIN)
        assert (
            initdb_bin
        ), f"Could not find '{INITDB_BIN}' or pg_config in system PATH. Make sure you have PostgreSQL installed."
        self.initdb_bin: str = initdb_bin

    def start(self) -> bool:
        """
        Start DB.

        :return: `True` if instance has been started or `False` if it could not start.
        """
        if self.running():
            return True

        try:
            old_wc = os.getcwd()
            self._create_db_path()
            assert self.db_path
            self._init_db()
            self._create_sockets_dir(self.db_path)

            os.chdir(self.db_path)
            args = [self.pg_ctl_bin, "start", "-D", ".", "-o", "-p " + str(self.port) + " -k " + "sockets", "-s"]
            process = subprocess.Popen(args, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
            process.communicate()
            return process.returncode == 0
        except Exception:
            LOGGER.exception("Failed to initialize the database.")
            return False
        finally:
            if old_wc is not None:
                os.chdir(old_wc)

    def _create_db_path(self) -> None:
        """Create the data directory to store the postgres data in"""
        if self.db_path:
            if not os.path.exists(self.db_path):
                os.mkdir(self.db_path)
            self._db_path_is_temporary = False
        else:
            self.db_path = tempfile.mkdtemp()
            self._db_path_is_temporary = True

    def _create_sockets_dir(self, parent_dir: str) -> str:
        sockets_dir = os.path.join(parent_dir, "sockets")
        if not os.path.exists(sockets_dir):
            os.mkdir(sockets_dir)
        return sockets_dir

    def _init_db(self) -> None:
        """Init the database if it is not a valid postgres data directory"""
        assert self.db_path
        if os.path.exists(os.path.join(self.db_path, "PG_VERSION")):
            return

        os.chmod(self.db_path, 0o700)
        args = [self.initdb_bin, "-D", self.db_path, "--auth-host", "trust", "-U", "postgres"]
        process = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        out, err = process.communicate()
        if process.returncode != 0:
            LOGGER.error("Failed to initialize db path")
            LOGGER.error("out: %s", out)
            LOGGER.error("err: %s", err)
            raise Exception("Failed to initialize db path.")

    def stop(self) -> None:
        if not self.running() or not self.db_path:
            return
        args = [self.pg_ctl_bin, "stop", "-D", self.db_path, "-m", "immediate", "-s"]
        process = subprocess.Popen(args, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
        process.communicate()
        if process.returncode != 0:
            raise Exception("Failed to stop embedded db.")
        if self._db_path_is_temporary and self.db_path:
            shutil.rmtree(self.db_path)
            self.db_path = None

    def running(self) -> bool:
        if self.db_path is None:
            return False
        args = [self.pg_ctl_bin, "status", "-D", self.db_path, "-s"]
        process = subprocess.Popen(args, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
        process.communicate()
        return process.returncode == 0



"""
    Copyright 2019 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import logging
import os
from collections import defaultdict
from typing import Any, Dict, Generic, List, Optional, TypeVar

import pkg_resources
import yaml

from inmanta import data
from inmanta.config import feature_file_config
from inmanta.data.model import ExtensionStatus
from inmanta.server.protocol import ServerSlice
from inmanta.stable_api import stable_api

LOGGER = logging.getLogger(__name__)


class InvalidSliceNameException(Exception):
    """This exception is raised when the name of the slice is not valid. For example when the extension and the slice name
    do not match.
    """


class InvalidFeature(Exception):
    """This exception is raised when a feature is defined in another slice than it reports or the feature is not defined
    at all.
    """


T = TypeVar("T")


class Feature(Generic[T]):
    """A feature offered by a slice"""

    def __init__(self, slice: str, name: str, description: str = "", default_value: Optional[T] = None) -> None:
        self._name: str = name
        self._slice: str = slice
        self._description = description
        self._default_value: Optional[T] = default_value

    @property
    def name(self) -> str:
        return self._name

    @property
    def slice(self) -> str:
        return self._slice

    @property
    def default_value(self) -> Optional[T]:
        return self._default_value

    def __str__(self) -> str:
        return f"{self._slice}:{self._name}"


class BoolFeature(Feature[bool]):
    """A feature that is on or off. When no value is given it is enabled."""

    def __init__(self, slice: str, name: str, description: str = "") -> None:
        super().__init__(slice, name, description, True)


class StringListFeature(Feature[List[str]]):
    """A feature that holds a list of allowed values. When the list contains "*" it matches everything."""

    def __init__(self, slice: str, name: str, description: str = "") -> None:
        super().__init__(slice, name, description, default_value=["*"])


class ProductMetadata:
    def __init__(self, product: str, edition: str, license: str, version: Optional[str]) -> None:
        self.product = product
        self.edition = edition
        self.license = license
        self.version = version


class FeatureManager:
    """This class allows to verify whether a feature should be enabled or not. This is determined based on a configuration
    file that is set with config.feature-file in the config. This feature file is a yaml with the following structure:

    .. code-block:: yaml

        slices:
            slice_name:
                feature_name: bool

    """

    def __init__(self) -> None:
        self._features: Dict[str, Dict[str, Feature[object]]] = defaultdict(lambda: {})
        self._feature_config: Dict[str, Dict[str, Any]] = self._load_feature_config()

    def get_features(self) -> List[Feature[object]]:
        return [feature for slice in self._features.values() for feature in slice.values()]

    def _load_feature_config(self) -> Dict[str, Dict[str, Any]]:
        feature_file = feature_file_config.get()
        if feature_file is None:
            return {}

        if not os.path.exists(feature_file):
            LOGGER.warning("Feature file %s configured but file does not exist.", feature_file)
            return {}

        with open(feature_file, encoding="utf-8") as fd:
            result = yaml.safe_load(fd)

        if "slices" in result:
            return result["slices"]
        return {}

    def get_product_metadata(self) -> ProductMetadata:
        return ProductMetadata(
            product="Inmanta Service Orchestrator",
            edition="Open Source Edition",
            license="Apache Software License 2",
            version=self._get_product_version(),
        )

    def _get_product_version(self) -> str:
        packages = ["inmanta-oss", "inmanta", "inmanta-core"]
        for package in packages:
            try:
                return pkg_resources.get_distribution(package).version
            except pkg_resources.DistributionNotFound:
                pass

        LOGGER.warning("Couldn't determine product version. Make sure inmanta is properly installed.")
        return "0.0.0"

    def add_slice(self, slice: ServerSlice) -> None:
        for feature in slice.define_features():
            if feature.slice != slice.name:
                raise InvalidFeature(
                    f"Feature {feature.name} defines slice {feature.slice} but is defined by slice {slice.name}"
                )

            self._features[feature.slice][feature.name] = feature
        slice.feature_manager = self

    def get_value(self, feature: Feature[T]) -> Any:
        """Get the value of a feature"""
        if feature.slice not in self._features or feature.name not in self._features[feature.slice]:
            raise InvalidFeature("Feature should be defined be slices at boot time.")

        if feature.slice not in self._feature_config or feature.name not in self._feature_config[feature.slice]:
            return feature.default_value
        return self._feature_config[feature.slice][feature.name]

    def enabled(self, feature: BoolFeature) -> bool:
        value = self.get_value(feature)
        return value

    def contains(self, feature: StringListFeature, item: str) -> bool:
        """Check if the value is contained in the list."""
        value = self.get_value(feature)
        if "*" in value:
            return True
        return item in value

    def stop(self) -> None:
        """Called when the server is stopped"""


@stable_api
class ApplicationContext:
    def __init__(self) -> None:
        self._slices: List[ServerSlice] = []
        self._feature_manager: Optional[FeatureManager] = None

    def register_slice(self, slice: ServerSlice) -> None:
        assert slice is not None
        self._slices.append(slice)

    def get_slices(self) -> List[ServerSlice]:
        return self._slices

    def set_feature_manager(self, feature_manager: FeatureManager) -> None:
        assert self._feature_manager is None
        self._feature_manager = feature_manager

    def get_feature_manager(self) -> FeatureManager:
        if self._feature_manager is None:
            self._feature_manager = FeatureManager()
        return self._feature_manager

    def get_product_metadata(self) -> ProductMetadata:
        return self.get_feature_manager().get_product_metadata()

    def get_extension_statuses(self) -> List[ExtensionStatus]:
        return ServerSlice.get_extension_statuses(list(self._slices))

    def register_environment_setting(self, setting: data.Setting) -> None:
        """
        Used by an Inmanta extension to register extension-specific environment settings.
        """
        data.Environment.register_setting(setting)

    def get_environment_settings(self) -> list[data.Setting]:
        """
        Returns the list of all available environment settings
        """
        return sorted(data.Environment._settings.values(), key=lambda x: x.name)



"""
    Copyright 2017 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

# pylint: disable-msg=W0613,R0201

import itertools
import logging
import uuid
from collections import abc
from itertools import chain
from typing import Dict, Iterator, List, Optional, Tuple, Union

import inmanta.ast.entity
import inmanta.ast.type as inmanta_type
import inmanta.execute.dataflow as dataflow
from inmanta.ast import (
    AmbiguousTypeException,
    AttributeReferenceAnchor,
    DuplicateException,
    InvalidCompilerState,
    Locatable,
    LocatableString,
    Location,
    Namespace,
    NotFoundException,
    Range,
    RuntimeException,
    TypeNotFoundException,
    TypeReferenceAnchor,
    TypingException,
)
from inmanta.ast.attribute import Attribute, RelationAttribute
from inmanta.ast.blocks import BasicBlock
from inmanta.ast.statements import (
    AttributeAssignmentLHS,
    ExpressionStatement,
    Literal,
    RawResumer,
    RequiresEmitStatement,
    StaticEagerPromise,
)
from inmanta.ast.statements.assign import GradualSetAttributeHelper, SetAttributeHelper
from inmanta.ast.variables import Reference
from inmanta.const import LOG_LEVEL_TRACE
from inmanta.execute.dataflow import DataflowGraph
from inmanta.execute.runtime import (
    ExecutionContext,
    ExecutionUnit,
    Instance,
    ListElementVariable,
    QueueScheduler,
    RawUnit,
    Resolver,
    ResultCollector,
    ResultVariable,
    VariableABC,
    VariableResolver,
    WrappedValueVariable,
)
from inmanta.execute.tracking import ImplementsTracker
from inmanta.execute.util import Unknown

try:
    from typing import TYPE_CHECKING
except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
    from inmanta.ast.entity import Entity, Implement  # noqa: F401

LOGGER = logging.getLogger(__name__)


class SubConstructor(ExpressionStatement):
    """
    This statement selects an implementation for a given object and
    imports the statements

    :ivar type: The specific entity type of an instance this subconstructor applies to, i.e. the actual instance type, not a
        supertype.
    """

    __slots__ = ("type", "location", "implements")

    def __init__(self, instance_type: "Entity", implements: "Implement") -> None:
        super().__init__()
        self.type = instance_type
        self.implements = implements
        self.location = self.implements.get_location()

    def normalize(self, *, lhs_attribute: Optional[AttributeAssignmentLHS] = None) -> None:
        # Only track promises for implementations when they get emitted, because of limitation of current static normalization
        # order: implementation blocks have not normalized at this point, so with the current mechanism we can't fetch eager
        # promises yet. Normalization order can not just be reversed because implementation bodies might contain constructor
        # calls (even for the same type), which would require this instance to be normalized first, resulting in a loop.
        self._own_eager_promises = []
        # injected_variables: Set[str] = {"self"}.union(self.type.get_all_attribute_names())
        # self._own_eager_promises = [
        #     # implementations live in the namespace's context rather than the constructor's context so for promises that cross
        #     # the boundary we translate references so that they are resolved correctly in any context wrapping the constructor
        #     dataclasses.replace(promise, instance=promise.instance.fully_qualified())
        #     for implementation in self.implements.implementations
        #     for promise in implementation.statements.get_eager_promises()
        #     if promise.get_root_variable() not in injected_variables
        # ]

    def requires_emit(self, resolver: Resolver, queue: QueueScheduler) -> Dict[object, VariableABC[object]]:
        requires: Dict[object, VariableABC[object]] = super().requires_emit(resolver, queue)
        try:
            resv = resolver.for_namespace(self.implements.constraint.namespace)
            requires.update(self.implements.constraint.requires_emit(resv, queue))
            return requires
        except NotFoundException as e:
            e.set_statement(self.implements)
            raise e

    def execute(self, requires: Dict[object, object], instance: Instance, queue: QueueScheduler) -> object:
        """
        Evaluate this statement
        """
        LOGGER.log(LOG_LEVEL_TRACE, "executing subconstructor for %s implement %s", self.type, self.implements.location)
        super().execute(requires, instance, queue)
        # this assertion is because the typing of this method is not correct
        # it should logically always hold, but we can't express this as types yet
        assert isinstance(instance, Instance)
        condition = self.implements.constraint.execute(requires, instance, queue)
        try:
            inmanta_type.Bool().validate(condition)
        except RuntimeException as e:
            e.set_statement(self.implements)
            e.msg = (
                "Invalid value `%s`: the condition for a conditional implementation can only be a boolean expression"
                % condition
            )
            raise e
        if not condition:
            return None

        myqueue = queue.for_tracker(ImplementsTracker(self, instance))

        implementations = self.implements.implementations

        for impl in implementations:
            if instance.add_implementation(impl):
                # generate a subscope/namespace for each loop
                xc = ExecutionContext(impl.statements, instance.for_namespace(impl.statements.namespace))
                xc.emit(myqueue)

        return None

    def pretty_print(self) -> str:
        return "implement %s using %s when %s" % (
            self.type,
            ",".join(i.name for i in self.implements.implementations),
            self.implements.constraint.pretty_print(),
        )

    def __str__(self) -> str:
        return self.pretty_print()

    def __repr__(self) -> str:
        return "SubConstructor(%s)" % self.type


class GradualFor(ResultCollector[object]):
    # this class might be unnecessary if receive-result is always called and exactly once

    def __init__(self, stmt: "For", resolver: Resolver, queue: QueueScheduler) -> None:
        self.resolver = resolver
        self.queue = queue
        self.stmt = stmt
        self.seen: set[int] = set()

    def receive_result(self, value: object, location: Location) -> bool:
        if id(value) in self.seen:
            return False
        self.seen.add(id(value))

        xc = ExecutionContext(self.stmt.module, self.resolver.for_namespace(self.stmt.module.namespace))
        loopvar = xc.lookup(self.stmt.loop_var)
        # this assertion is because the typing of this method is not correct
        # it should logically always hold, but we can't express this as types yet
        assert isinstance(loopvar, ResultVariable)
        loopvar.set_value(value, self.stmt.location)
        xc.emit(self.queue)
        return False


class For(RequiresEmitStatement):
    """
    A for loop
    """

    __slots__ = ("base", "loop_var", "loop_var_loc", "module")

    def __init__(self, variable: ExpressionStatement, loop_var: LocatableString, module: BasicBlock) -> None:
        super().__init__()
        self.base: ExpressionStatement = variable
        self.loop_var = str(loop_var)
        self.loop_var_loc = loop_var.get_location()
        self.module = module
        self.anchors.extend(module.get_anchors())
        self.anchors.extend(variable.get_anchors())

    def __repr__(self) -> str:
        return "For(%s)" % self.loop_var

    def normalize(self) -> None:
        self.base.normalize()
        # self.loop_var.normalize(resolver)
        self.module.normalize()
        self.module.add_var(self.loop_var, self)
        self._own_eager_promises = self.module.get_eager_promises()

    def get_all_eager_promises(self) -> Iterator["StaticEagerPromise"]:
        return chain(super().get_all_eager_promises(), self.base.get_all_eager_promises())

    def requires_emit(self, resolver: Resolver, queue: QueueScheduler) -> dict[object, VariableABC[object]]:
        requires: Dict[object, VariableABC[object]] = super().requires_emit(resolver, queue)

        # pass context via requires!
        helper = GradualFor(self, resolver, queue)
        requires[self] = WrappedValueVariable(helper)

        requires.update(self.base.requires_emit_gradual(resolver, queue, helper))

        return requires

    def execute(self, requires: Dict[object, object], resolver: Resolver, queue: QueueScheduler) -> object:
        """
        Evaluate this statement.
        """
        super().execute(requires, resolver, queue)
        var = self.base.execute(requires, resolver, queue)

        if isinstance(var, Unknown):
            return None

        if not isinstance(var, list):
            raise TypingException(self, "A for loop can only be applied to lists and relations")

        helper = requires[self]
        assert isinstance(helper, GradualFor)

        for loop_var in var:
            # generate a subscope/namespace for each loop
            helper.receive_result(loop_var, self.location)

        return None

    def nested_blocks(self) -> Iterator["BasicBlock"]:
        yield self.module


class ListComprehension(RawResumer, ExpressionStatement):
    """
    A list comprehension expression, e.g. `["hello {{world}}" for world in worlds if world != "exclude"]`.
    """

    __slots__ = ("loop_var", "value_expression", "iterable", "guard")

    def __init__(
        self,
        value_expression: ExpressionStatement,
        loop_var: LocatableString,
        iterable: ExpressionStatement,
        guard: Optional[ExpressionStatement] = None,
    ) -> None:
        super().__init__()
        self.value_expression: ExpressionStatement = value_expression
        self.loop_var: LocatableString = loop_var
        self.iterable: ExpressionStatement = iterable
        self.guard: Optional[ExpressionStatement] = guard
        self.anchors.extend(
            itertools.chain(
                self.value_expression.get_anchors(),
                self.iterable.get_anchors(),
                (self.guard.get_anchors() if self.guard is not None else ()),
            )
        )

    def normalize(self, *, lhs_attribute: Optional[AttributeAssignmentLHS] = None) -> None:
        self.value_expression.normalize(lhs_attribute=lhs_attribute)
        self.iterable.normalize()
        if self.guard is not None:
            self.guard.normalize()

    def requires(self) -> list[str]:
        # exclude loop var, unless it shadows an occurrence in iterable
        return list(set(self.value_expression.requires()) - {str(self.loop_var)} | set(self.iterable.requires()))

    def requires_emit(
        self, resolver: Resolver, queue: QueueScheduler, *, lhs: Optional[ResultCollector[object]] = None
    ) -> dict[object, VariableABC[object]]:
        """
        Sets up gradual or non-gradual execution (depending on the lhs) of the list comprehension. Additionally sets up a
        resumer for when the iterable is complete. Returns as requires the (gradual) helper's result variable, which will be
        frozen by the resumer.

        Flow:
            `requires_emit()` -> set up helper, schedule resume and execute
            -> helper gradual execution (if lhs is not None)
            -> wait for iterable completion -> `resume()` -> finalize helper input
            -> wait for helper completion -> `execute()` -> return complete value
        """
        base_requires: dict[object, VariableABC[object]] = super().requires_emit(resolver, queue)

        # set up gradual execution
        collector_helper: ListComprehensionCollector = ListComprehensionCollector(
            statement=self,
            resolver=resolver,
            queue=queue,
            lhs=lhs,
        )
        self.copy_location(collector_helper)
        iterable_requires: dict[object, VariableABC[object]] = (
            # use helper strictly non-gradually when in a non-gradual context
            # => propagates progress potential to iterable expression's requires
            self.iterable.requires_emit(resolver, queue)
            if lhs is None
            else self.iterable.requires_emit_gradual(resolver, queue, collector_helper)
        )

        # non-gradual mode / finishing up: resume as soon as the iterable can be executed
        # pass helper to the resumer via the requires object
        wrapped_helper: VariableABC[ListComprehensionCollector] = WrappedValueVariable(collector_helper)
        requires: dict[object, VariableABC[object]] = base_requires | iterable_requires | {self: wrapped_helper}
        RawUnit(queue, resolver, requires, resumer=self)

        # Wait for resumer and helper to populate result.
        # No need to wait for iterable requires explicitly because resumer already does
        return base_requires | {self: collector_helper.final_result}

    def requires_emit_gradual(
        self, resolver: Resolver, queue: QueueScheduler, resultcollector: ResultCollector[object]
    ) -> dict[object, VariableABC[object]]:
        return self.requires_emit(resolver, queue, lhs=resultcollector)

    def resume(self, requires: dict[object, VariableABC[object]], resolver: Resolver, queue: QueueScheduler) -> None:
        """
        Resume when the iterable is fully ready for execution. Populates the helper's result variable and signals that no more
        values will be sent.
        """
        # fetch helper, passed via the requires object
        collector_helper: object = requires[self].get_value()
        assert isinstance(collector_helper, ListComprehensionCollector)

        # execute iterable so we have all values
        iterable: object = self.iterable.execute({k: v.get_value() for k, v in requires.items()}, resolver, queue)

        # indicate to helper that we're done
        if isinstance(iterable, Unknown):
            collector_helper.set_unknown()
        elif not isinstance(iterable, list):
            raise TypingException(
                self, f"A list comprehension can only be applied to lists and relations, got {type(iterable).__name__}"
            )
        else:
            collector_helper.complete(iterable, resolver, queue)

    def execute(self, requires: dict[object, object], resolver: Resolver, queue: QueueScheduler) -> object:
        super().execute(requires, resolver, queue)
        # at this point the resumer signalled the helper we were done and the helper waited for all value expressions
        # => just fetch the result
        return requires[self]

    def execute_direct(self, requires: abc.Mapping[str, object]) -> Union[list[object], Unknown]:
        iterable: object = self.iterable.execute_direct(requires)
        if isinstance(iterable, Unknown):
            return Unknown(self)
        elif not isinstance(iterable, list):
            raise TypingException(
                self,
                f"A list comprehension in a direct execute context can only be applied to lists, got {type(iterable).__name__}",
            )

        def process(element: object) -> Optional[object]:
            """
            Execute the list comprehension for a single element of the iterable. Evaluates the guard expression if there is
            any and executes the value expression if the guard passes. Returns the result of the value expression if the
            guard passed, otherwise returns None.
            """
            extended_requires: abc.Mapping[str, object] = requires | {str(self.loop_var): element}

            guard_passed: bool
            if self.guard is None:
                guard_passed = True
            else:
                guard_value: object = self.guard.execute_direct(extended_requires)
                if isinstance(iterable, Unknown):
                    return Unknown(self)
                if not isinstance(guard_value, bool):
                    raise TypingException(
                        self,
                        (
                            f"Invalid value `{guard_value}`:"
                            " the guard condition for a list comprehension must be a boolean expression"
                        ),
                    )
                guard_passed = guard_value

            return self.value_expression.execute_direct(extended_requires) if guard_passed else None

        return [result for element in iterable if (result := process(element)) is not None]

    def pretty_print(self) -> str:
        return "[%s for %s in %s%s]" % (
            self.value_expression.pretty_print(),
            self.loop_var,
            self.iterable.pretty_print(),
            f" if {self.guard.pretty_print()}" if self.guard is not None else "",
        )

    def __str__(self) -> str:
        return self.pretty_print()

    def __repr__(self) -> str:
        return "ListComprehension(value_expression=%s, loop_var=%s, iterable=%s, guard=%s)" % (
            repr(self.value_expression),
            repr(self.loop_var),
            repr(self.iterable),
            repr(self.guard),
        )


LIST_COMPREHENSION_GUARDED = object()
"""
Artificial value used in the else branch of the list comprehension guard's conditional expression. Indicates that the value
expression for an element should not be executed because the element was filtered by the guard.
"""


class ListComprehensionCollector(RawResumer, ResultCollector[object]):
    """
    Result collector (gradual or otherwise) for the list comprehension statement. When it receives a
    result, it sets up appropriate (gradual or otherwise) execution for the value expression, with the lhs as final result
    collector. Finally, it collects all final results in its own result variable for non-gradual execution.

    This class has a gradual mode and a non-gradual mode depending on whether `lhs` is set. In non-gradual mode (lhs is None)
    it must not receive any gradual results. This allows the class to maintain ordering in non-gradual mode (if elements are
    received gradually each they arrive as soon as they can be resolved).

    Expects to receive each result once, either gradually or at execution-time. Clients should indicate completion by calling
    `complete()` or `set_unknown()`, after which no more values may be provided. This instance's `final_result` attribute
    will contain the final result. It will be only be set after the client declares completion, but not necessarily immediately
    after (we may have to wait for value expressions' requires).
    """

    __slots__ = ("statement", "resolver", "queue", "lhs", "_results", "_complete", "final_result")

    def __init__(
        self,
        statement: ListComprehension,
        resolver: Resolver,
        queue: QueueScheduler,
        lhs: Optional[ResultCollector[object]] = None,
    ) -> None:
        self.statement: ListComprehension = statement
        self.resolver: Resolver = resolver
        self.queue: QueueScheduler = queue
        self.lhs: Optional[ResultCollector[object]] = lhs
        # separately collect results for the value expression for each element of the iterable so we can wait for
        # each element's value expression to complete.
        self._results: list[VariableABC[object]] = []
        # have we received all elements?
        self._complete: bool = False
        # collector for the final result
        self.final_result: ResultVariable[object] = ResultVariable()

    def receive_result(self, value: object, location: Location) -> bool:
        """
        Receive an single element from the iterable and schedule execution of the associated value expression.
        """
        if self._complete:
            raise InvalidCompilerState(self, "list comprehension helper received gradual result after it was declared complete")

        # Use special result variable for a pre-known value with listener capabilities. Using a plain `ResultVariable` would
        # break gradual execution (e.g. `Reference.requires_emit_gradual` registers self.lhs as listener)
        value_wrapper: ListElementVariable[object] = ListElementVariable(value, location)
        value_resolver: VariableResolver = VariableResolver(
            parent=self.resolver,
            name=str(self.statement.loop_var),
            variable=value_wrapper,
        )

        result_variable: ResultVariable[object] = ResultVariable()
        self._results.append(result_variable)

        # execute the value expression and the guard
        guarded_expression: ExpressionStatement
        if self.statement.guard is None:
            guarded_expression = self.statement.value_expression
        else:
            else_expression: ExpressionStatement = Literal(LIST_COMPREHENSION_GUARDED)
            guarded_expression = ConditionalExpression(
                condition=self.statement.guard,
                if_expression=self.statement.value_expression,
                else_expression=Literal(LIST_COMPREHENSION_GUARDED),
            )
            self.statement.copy_location(else_expression)
            self.statement.copy_location(guarded_expression)

        requires: dict[object, VariableABC[object]] = (
            guarded_expression.requires_emit(value_resolver, self.queue)
            if self.lhs is None
            else guarded_expression.requires_emit_gradual(value_resolver, self.queue, self.lhs)
        )
        ExecutionUnit(
            self.queue,
            value_resolver,
            result_variable,
            requires,
            guarded_expression,
        )

        return False

    def set_unknown(self) -> None:
        """
        Set the final result to be an unknown. No elements should be gradually received in this case.
        Mutually exclusive with `complete`.
        """
        if self._results:
            raise InvalidCompilerState(
                self, "list comprehension helper got set_unknown after some (known) elements where received"
            )
        self.final_result.set_value(Unknown(self.statement), self.statement.location)

    def complete(self, all_values: abc.Sequence[object], resolver: Resolver, queue: QueueScheduler) -> None:
        """
        Indicate that all results have been received. No further calls to `receive_result` should be done after this.
        Mutually exclusive with `set_unknown`.
        """
        if self._results:
            # We should only have received previous results in gradual mode, if the
            if self.lhs is None:
                raise InvalidCompilerState(self, "list comprehension helper received gradual results in non-gradual mode")
            if len(self._results) != len(all_values):
                raise InvalidCompilerState(self, "list comprehension helper received some but not all values gradually")
        else:
            for value in all_values:
                self.receive_result(value, location=self.statement.location)

        RawUnit(queue, resolver, dict(enumerate(self._results)), resumer=self)

    def resume(self, requires: dict[object, VariableABC[object]], resolver: Resolver, queue: QueueScheduler) -> None:
        def get(variable: VariableABC[object]) -> Optional[abc.Sequence[object]]:
            value: object = variable.get_value()
            return None if value is LIST_COMPREHENSION_GUARDED else value if isinstance(value, list) else [value]

        # collect all element value expressions' results and write them to the final result variable
        self.final_result.set_value(
            list(
                itertools.chain.from_iterable(value for variable in requires.values() if (value := get(variable)) is not None)
            ),
            self.statement.location,
        )


class If(RequiresEmitStatement):
    """
    An if Statement
    """

    __slots__ = ("condition", "if_branch", "else_branch")

    def __init__(self, condition: ExpressionStatement, if_branch: BasicBlock, else_branch: BasicBlock) -> None:
        super().__init__()
        self.condition: ExpressionStatement = condition
        self.if_branch: BasicBlock = if_branch
        self.else_branch: BasicBlock = else_branch
        self.anchors.extend(condition.get_anchors())
        self.anchors.extend(if_branch.get_anchors())
        self.anchors.extend(else_branch.get_anchors())

    def __repr__(self) -> str:
        return "If"

    def normalize(self) -> None:
        self.condition.normalize()
        self.if_branch.normalize()
        self.else_branch.normalize()
        self._own_eager_promises = [*self.if_branch.get_eager_promises(), *self.else_branch.get_eager_promises()]

    def get_all_eager_promises(self) -> Iterator["StaticEagerPromise"]:
        return chain(super().get_all_eager_promises(), self.condition.get_all_eager_promises())

    def requires_emit(self, resolver: Resolver, queue: QueueScheduler) -> Dict[object, VariableABC[object]]:
        requires: Dict[object, VariableABC[object]] = super().requires_emit(resolver, queue)
        requires.update(self.condition.requires_emit(resolver, queue))
        return requires

    def execute(self, requires: Dict[object, object], resolver: Resolver, queue: QueueScheduler) -> object:
        """
        Evaluate this statement.
        """
        super().execute(requires, resolver, queue)
        cond: object = self.condition.execute(requires, resolver, queue)
        if isinstance(cond, Unknown):
            return None
        try:
            inmanta_type.Bool().validate(cond)
        except RuntimeException as e:
            e.set_statement(self)
            e.msg = "Invalid value `%s`: the condition for an if statement can only be a boolean expression" % cond
            raise e
        # schedule appropriate branch body
        branch: BasicBlock = self.if_branch if cond else self.else_branch
        xc = ExecutionContext(branch, resolver.for_namespace(branch.namespace))
        xc.emit(queue)
        return None

    def nested_blocks(self) -> Iterator["BasicBlock"]:
        yield self.if_branch
        yield self.else_branch


class ConditionalExpression(ExpressionStatement):
    """
    A conditional expression similar to Python's `x if c else y`.
    """

    __slots__ = ("condition", "if_expression", "else_expression")

    def __init__(
        self, condition: ExpressionStatement, if_expression: ExpressionStatement, else_expression: ExpressionStatement
    ) -> None:
        super().__init__()
        self.condition: ExpressionStatement = condition
        self.if_expression: ExpressionStatement = if_expression
        self.else_expression: ExpressionStatement = else_expression
        self.anchors.extend(condition.get_anchors())
        self.anchors.extend(if_expression.get_anchors())
        self.anchors.extend(else_expression.get_anchors())

    def normalize(self, *, lhs_attribute: Optional[AttributeAssignmentLHS] = None) -> None:
        self.condition.normalize()
        # pass on lhs_attribute to branches
        self.if_expression.normalize(lhs_attribute=lhs_attribute)
        self.else_expression.normalize(lhs_attribute=lhs_attribute)
        self._own_eager_promises = [
            *self.if_expression.get_all_eager_promises(),
            *self.else_expression.get_all_eager_promises(),
        ]

    def get_all_eager_promises(self) -> Iterator["StaticEagerPromise"]:
        return chain(super().get_all_eager_promises(), self.condition.get_all_eager_promises())

    def requires(self) -> List[str]:
        return list(chain.from_iterable(sub.requires() for sub in [self.condition, self.if_expression, self.else_expression]))

    def requires_emit(
        self, resolver: Resolver, queue: QueueScheduler, *, lhs: Optional[ResultCollector[object]] = None
    ) -> dict[object, VariableABC[object]]:
        requires: Dict[object, VariableABC[object]] = super().requires_emit(resolver, queue)

        # This ResultVariable will receive the result of this expression
        result: ResultVariable[object] = ResultVariable()

        # Schedule execution to resume when the condition can be executed
        resumer: RawResumer = ConditionalExpressionResumer(self, result, lhs=lhs)
        self.copy_location(resumer)
        RawUnit(queue, resolver, self.condition.requires_emit(resolver, queue), resumer)

        # Wait for the result variable to be populated
        requires[self] = result
        return requires

    def requires_emit_gradual(
        self, resolver: Resolver, queue: QueueScheduler, resultcollector: ResultCollector[object]
    ) -> dict[object, VariableABC[object]]:
        return self.requires_emit(resolver, queue, lhs=resultcollector)

    def execute(self, requires: Dict[object, object], resolver: Resolver, queue: QueueScheduler) -> object:
        super().execute(requires, resolver, queue)
        return requires[self]

    def execute_direct(self, requires: abc.Mapping[str, object]) -> object:
        condition_value: object = self.condition.execute_direct(requires)
        if isinstance(condition_value, Unknown):
            return Unknown(self)
        if not isinstance(condition_value, bool):
            raise RuntimeException(
                self, "Invalid value `%s`: the condition for a conditional expression must be a boolean expression"
            )
        return (self.if_expression if condition_value else self.else_expression).execute_direct(requires)

    def pretty_print(self) -> str:
        return "%s ? %s : %s" % (
            self.condition.pretty_print(),
            self.if_expression.pretty_print(),
            self.else_expression.pretty_print(),
        )

    def __repr__(self) -> str:
        return "%s ? %s : %s" % (self.condition, self.if_expression, self.else_expression)


class ConditionalExpressionResumer(RawResumer):
    __slots__ = ("expression", "condition_value", "result", "lhs")

    def __init__(
        self, expression: ConditionalExpression, result: ResultVariable, *, lhs: Optional[ResultCollector[object]] = None
    ) -> None:
        super().__init__()
        self.expression: ConditionalExpression = expression
        self.condition_value: Optional[bool] = None
        self.result: ResultVariable = result
        self.lhs: Optional[ResultCollector[object]] = lhs

    def resume(self, requires: Dict[object, VariableABC[object]], resolver: Resolver, queue: QueueScheduler) -> None:
        if self.condition_value is None:
            condition_value: object = self.expression.condition.execute(
                {k: v.get_value() for k, v in requires.items()}, resolver, queue
            )
            if isinstance(condition_value, Unknown):
                self.result.set_value(Unknown(self), self.location)
                return
            if not isinstance(condition_value, bool):
                raise RuntimeException(
                    self, "Invalid value `%s`: the condition for a conditional expression must be a boolean expression"
                )
            self.condition_value = condition_value

            # Schedule execution of appropriate subexpression
            subexpression: ExpressionStatement = (
                self.expression.if_expression if self.condition_value else self.expression.else_expression
            )
            subexpression_requires: dict[object, VariableABC[object]] = (
                subexpression.requires_emit(resolver, queue)
                if self.lhs is None
                else subexpression.requires_emit_gradual(resolver, queue, self.lhs)
            )
            RawUnit(
                queue,
                resolver,
                subexpression_requires,
                self,
            )
        else:
            value: object = (
                self.expression.if_expression if self.condition_value else self.expression.else_expression
            ).execute({k: v.get_value() for k, v in requires.items()}, resolver, queue)
            self.result.set_value(value, self.location)


class IndexAttributeMissingInConstructorException(TypingException):
    """
    Raised when an index attribute was not set in the constructor call for an entity.
    """

    def __init__(self, stmt: Optional[Locatable], entity: "Entity", unset_attributes: abc.Sequence[str]):
        if not unset_attributes:
            raise Exception("Argument `unset_attributes` should contain at least one element")
        error_message = self._get_error_message(entity, unset_attributes)
        super(IndexAttributeMissingInConstructorException, self).__init__(stmt, error_message)

    def _get_error_message(self, entity: "Entity", unset_attributes: abc.Sequence[str]) -> str:
        exc_message = "Invalid Constructor call:"
        for attribute_name in unset_attributes:
            attribute: Optional[Attribute] = entity.get_attribute(attribute_name)
            assert attribute is not None  # Make mypy happy
            attribute_kind = "relation" if isinstance(attribute, RelationAttribute) else "attribute"
            exc_message += (
                f"\n\t* Missing {attribute_kind} '{attribute.name}'. "
                f"The {attribute_kind} {entity.get_full_name()}.{attribute.name} is part of an index."
            )
        return exc_message


class Constructor(ExpressionStatement):
    """
    This class represents the usage of a constructor to create a new object.

    :param class_type: The type of the object that is created by this
        constructor call.
    """

    __slots__ = (
        "class_type",
        "__attributes",
        "__attribute_locations",
        "__wrapped_kwarg_attributes",
        "location",
        "type",
        "_self_ref",
        "_lhs_attribute",
        "_required_dynamic_args",
        "_direct_attributes",
        "_indirect_attributes",
    )

    def __init__(
        self,
        class_type: LocatableString,
        attributes: List[Tuple[LocatableString, ExpressionStatement]],
        wrapped_kwargs: List["WrappedKwargs"],
        location: Location,
        namespace: Namespace,
    ) -> None:
        super().__init__()
        self.class_type = class_type
        self.__attributes = {}  # type: Dict[str,ExpressionStatement]
        self.__attribute_locations: Dict[str, LocatableString] = {}
        self.__wrapped_kwarg_attributes: List[WrappedKwargs] = wrapped_kwargs
        self.location = location
        self.namespace = namespace
        self.anchors.append(TypeReferenceAnchor(namespace, class_type))
        for a in attributes:
            self.add_attribute(a[0], a[1])
        self.type: Optional["Entity"] = None
        self._self_ref: "Reference" = Reference(
            LocatableString(str(uuid.uuid4()), Range("__internal__", 1, 1, 1, 1), -1, self.namespace)
        )
        self._lhs_attribute: Optional[AttributeAssignmentLHS] = None
        self._required_dynamic_args: list[str] = []  # index attributes required from kwargs or lhs_attribute

        self._direct_attributes = {}  # type: Dict[str,ExpressionStatement]
        self._indirect_attributes = {}  # type: Dict[str,ExpressionStatement]

    def pretty_print(self) -> str:
        return "%s(%s)" % (
            self.class_type,
            ",".join(
                chain(
                    ("%s=%s" % (k, v.pretty_print()) for k, v in self.attributes.items()),
                    ("**%s" % kwargs.pretty_print() for kwargs in self.wrapped_kwargs),
                )
            ),
        )

    def _normalize_rhs(self, index_attributes: abc.Set[str]) -> None:
        assert self.type is not None  # Make mypy happy
        for k, v in self.__attributes.items():
            attr = self.type.get_attribute(k)
            if attr is None:
                raise TypingException(
                    self.__attribute_locations[k], "no attribute %s on type %s" % (k, self.type.get_full_name())
                )
            type_hint = attr.get_type().get_base_type()
            # don't notify the rhs for index attributes because it won't be able to resolve the reference
            # (index attributes need to be resolved before the instance can be constructed)
            v.normalize(
                lhs_attribute=AttributeAssignmentLHS(self._self_ref, k, type_hint) if k not in index_attributes else None
            )
        for wrapped_kwargs in self.wrapped_kwargs:
            wrapped_kwargs.normalize()

    def normalize(self, *, lhs_attribute: Optional[AttributeAssignmentLHS] = None) -> None:
        self.type = self._resolve_type(lhs_attribute)

        inindex: abc.MutableSet[str] = set()

        all_attributes = dict(self.type.get_default_values())
        all_attributes.update(self.__attributes)

        # now check that all variables that have indexes on them, are already
        # defined and add the instance to the index
        for index in self.type.get_indices():
            for attr in index:
                if attr not in all_attributes:
                    self._required_dynamic_args.append(attr)
                    continue
                inindex.add(attr)

        if self._required_dynamic_args:
            # Limit dynamic compile-time overhead: ignore lhs if this constructor doesn't need it for instantiation.
            # Concretely, only store it if not all index attributes are explicitly set in the constructor.
            self._lhs_attribute = lhs_attribute
            # raise an exception if there are more required dynamic arguments than could be provided by the kwargs and
            # lhs attribute. If this passes but the kwargs and/or lhs attribute don't in fact provide the required arguments,
            # an exception is raised during execution.
            if not self.wrapped_kwargs and (self._lhs_attribute is None or len(self._required_dynamic_args) > 1):
                raise IndexAttributeMissingInConstructorException(self, self.type, self._required_dynamic_args)

        self._normalize_rhs(inindex)

        for k, v in all_attributes.items():
            attribute = self.type.get_attribute(k)
            if attribute is None:
                raise TypingException(
                    self.__attribute_locations[k], "no attribute %s on type %s" % (k, self.type.get_full_name())
                )
            if k not in inindex:
                self._indirect_attributes[k] = v
            else:
                self._direct_attributes[k] = v

        self._own_eager_promises = list(
            chain.from_iterable(subconstructor.get_all_eager_promises() for subconstructor in self.type.get_sub_constructor())
        )

    def _resolve_type(self, lhs_attribute: Optional[AttributeAssignmentLHS]) -> "Entity":
        """Type hint handling"""

        # First normal resolution
        resolver_failure: Optional[TypeNotFoundException] = None
        local_type: "Optional[Entity]" = None
        try:
            tp = self.namespace.get_type(self.class_type)
            assert isinstance(
                tp, inmanta.ast.entity.Entity
            ), "Should not happen because all entity types start with a capital letter"
            local_type = tp
        except TypeNotFoundException as e:
            resolver_failure = e

        # Do we have hint context?
        # We only work with unqualified names for hinting
        if lhs_attribute is not None and lhs_attribute.type_hint is not None:
            # We can do type hinting here
            type_hint = lhs_attribute.type_hint
            if not isinstance(type_hint, inmanta.ast.entity.Entity):
                # This is a type error, we are a constructor for an entity but we should not be!
                raise TypingException(
                    self,
                    f"Can not assign a value of type {self.class_type} "
                    f"to a variable of type {str(lhs_attribute.type_hint)}",
                )
            elif local_type is not None:
                # always prefer local type, raise an exception if it is of an incorrect type
                if not type_hint.is_subclass(local_type, strict=False):
                    raise TypingException(
                        self,
                        f"Can not assign a value of type {str(local_type)} "
                        f"to a variable of type {str(lhs_attribute.type_hint)}",
                    )
                return local_type
            elif "::" not in str(self.class_type):
                # Consider the hint type
                base_types = [type_hint]
                # Find all correct types with the matching unqualified name
                candidates = {
                    entity
                    for entity in chain(base_types, type_hint.get_all_child_entities())
                    if entity.name == str(self.class_type)
                }
                if len(candidates) > 1:
                    # To many options, inheritance may cause this to break a working model due to dependency update
                    raise AmbiguousTypeException(self.class_type, list(candidates))
                elif len(candidates) == 1:
                    # One, nice
                    return next(iter(candidates))
        elif local_type is not None:
            return local_type

        # No matching types found: pretend nothing happened, reraise original exception
        assert resolver_failure is not None  # make mypy happy
        raise resolver_failure

    def get_all_eager_promises(self) -> Iterator["StaticEagerPromise"]:
        return chain(
            super().get_all_eager_promises(),
            *(subexpr.get_all_eager_promises() for subexpr in chain(self.attributes.values(), self.wrapped_kwargs)),
        )

    def requires(self) -> List[str]:
        out = [req for (k, v) in self.__attributes.items() for req in v.requires()]
        out.extend(req for kwargs in self.__wrapped_kwarg_attributes for req in kwargs.requires())
        out.extend(req for (k, v) in self.get_default_values().items() for req in v.requires())
        return out

    def requires_emit(self, resolver: Resolver, queue: QueueScheduler) -> Dict[object, VariableABC[object]]:
        requires: Dict[object, VariableABC[object]] = super().requires_emit(resolver, queue)
        # direct
        direct = [x for x in self._direct_attributes.items()]

        direct_requires = {rk: rv for (k, v) in direct for (rk, rv) in v.requires_emit(resolver, queue).items()}
        direct_requires.update(
            {rk: rv for kwargs in self.__wrapped_kwarg_attributes for (rk, rv) in kwargs.requires_emit(resolver, queue).items()}
        )
        if self._lhs_attribute is not None:
            direct_requires.update(
                # if lhs_attribute is set, it is likely required for construction (only exception is if it is in kwargs)
                self._lhs_attribute.instance.requires_emit(resolver, queue)
            )
        LOGGER.log(
            LOG_LEVEL_TRACE, "emitting constructor for %s at %s with %s", self.class_type, self.location, direct_requires
        )
        requires.update(direct_requires)

        graph: Optional[DataflowGraph] = resolver.dataflow_graph
        if graph is not None:
            node: dataflow.InstanceNodeReference = self._register_dataflow_node(graph)
            # TODO: also add wrapped_kwargs
            for k, v in chain(self._direct_attributes.items(), self._indirect_attributes.items()):
                node.assign_attribute(k, v.get_dataflow_node(graph), self, graph)

        return requires

    def _collect_required_dynamic_arguments(
        self, requires: Dict[object, object], resolver: Resolver, queue: QueueScheduler
    ) -> abc.Mapping[str, object]:
        """
        Part of the execute flow: returns values for kwargs and the inverse relation derived from the lhs for which this
        constructor is the rhs, if appliccable.
        """
        type_class = self.type
        assert type_class

        # kwargs
        kwarg_attrs: dict[str, object] = {}
        for kwargs in self.wrapped_kwargs:
            for k, v in kwargs.execute(requires, resolver, queue):
                if k in self.attributes or k in kwarg_attrs:
                    raise RuntimeException(
                        self, "The attribute %s is set twice in the constructor call of %s." % (k, self.class_type)
                    )
                attribute = type_class.get_attribute(k)
                if attribute is None:
                    raise TypingException(self, "no attribute %s on type %s" % (k, type_class.get_full_name()))
                kwarg_attrs[k] = v

        lhs_inverse_assignment: Optional[tuple[str, object]] = None
        # add inverse relation if it is part of an index
        if self._lhs_attribute is not None:
            lhs_instance: object = self._lhs_attribute.instance.execute(requires, resolver, queue)
            if not isinstance(lhs_instance, Instance):
                # bug in internal implementation
                raise Exception("Invalid state: received lhs_attribute that is not an instance")
            lhs_attribute: Optional[Attribute] = lhs_instance.get_type().get_attribute(self._lhs_attribute.attribute)
            if not isinstance(lhs_attribute, RelationAttribute):
                # bug in the model
                raise RuntimeException(
                    self,
                    (
                        f"Attempting to assign constructor of type {type_class} to attribute that is not a relation attribute:"
                        f" {lhs_attribute} on {lhs_instance}"
                    ),
                )
            inverse: Optional[RelationAttribute] = lhs_attribute.end
            if (
                inverse is not None
                and inverse.name not in self._direct_attributes
                # in case of a double set, prefer kwargs: double set will be raised when the bidirectional relation is set by
                # the LHS
                and inverse.name not in kwarg_attrs
                and inverse.name in chain.from_iterable(type_class.get_indices())
                and (inverse.entity == type_class or type_class.is_parent(inverse.entity))
            ):
                lhs_inverse_assignment = (inverse.name, lhs_instance)

        late_args = {**dict([lhs_inverse_assignment] if lhs_inverse_assignment is not None else []), **kwarg_attrs}
        missing_attrs: abc.Sequence[str] = [attr for attr in self._required_dynamic_args if attr not in late_args]
        if missing_attrs:
            raise IndexAttributeMissingInConstructorException(self, type_class, missing_attrs)
        return late_args

    def execute(self, requires: Dict[object, object], resolver: Resolver, queue: QueueScheduler) -> Instance:
        """
        Evaluate this statement.
        """
        LOGGER.log(LOG_LEVEL_TRACE, "executing constructor for %s at %s", self.class_type, self.location)
        super().execute(requires, resolver, queue)

        # the type to construct
        type_class = self.type
        assert type_class

        # kwargs and implicit inverse from lhs
        late_args: abc.Mapping[str, object] = self._collect_required_dynamic_arguments(requires, resolver, queue)

        # Schedule all direct attributes for direct execution. The kwarg keys and the direct_attributes keys are disjoint
        # because a RuntimeException is raised above when they are not.
        direct_attributes: Dict[str, object] = {
            k: v.execute(requires, resolver, queue) for (k, v) in self._direct_attributes.items()
        }
        direct_attributes.update(late_args)

        # Override defaults with kwargs. The kwarg keys and the indirect_attributes keys are disjoint because a RuntimeException
        # is raised above when they are not.
        indirect_attributes: Dict[str, ExpressionStatement] = {
            k: v for k, v in self._indirect_attributes.items() if k not in late_args
        }

        # check if the instance already exists in the index (if there is one)
        instances: List[Instance] = []
        # register any potential index collision
        collisions: abc.MutableMapping[tuple[str, ...], Instance] = {}
        for index in type_class.get_indices():
            params = []
            for attr in index:
                params.append((attr, direct_attributes[attr]))

            obj: Optional[Instance] = type_class.lookup_index(params, self)

            if obj is not None:
                if obj.get_type() != type_class:
                    raise DuplicateException(self, obj, "Type found in index is not an exact match")
                instances.append(obj)
                collisions[tuple(index)] = obj

        object_instance: Instance
        graph: Optional[DataflowGraph] = resolver.dataflow_graph
        if len(instances) > 0:
            if graph is not None:
                graph.add_index_match(
                    chain(
                        [self.get_dataflow_node(graph)],
                        (i.instance_node for i in instances if i.instance_node is not None),
                    )
                )

            # ensure that instances are all the same objects
            first = instances[0]
            for i in instances[1:]:
                if i != first:
                    raise IndexCollisionException(
                        msg=("Inconsistent indexes detected!\n"),
                        constructor=self,
                        collisions=collisions,
                    )

            object_instance = first
            self.copy_location(object_instance)
            for k, v in direct_attributes.items():
                object_instance.set_attribute(k, v, self.location)
        else:
            # create the instance
            object_instance = type_class.get_instance(
                direct_attributes, resolver, queue, self.location, self.get_dataflow_node(graph) if graph is not None else None
            )

        # deferred execution for indirect attributes
        # inject implicit reference to this instance so attributes can resolve the lhs_attribute we promised in _normalize_rhs
        self_var: ResultVariable[Instance] = ResultVariable()
        self_var.set_value(object_instance, self.location)
        self_resolver: VariableResolver = VariableResolver(resolver, self._self_ref.name, self_var)
        for attributename, valueexpression in indirect_attributes.items():
            var = object_instance.get_attribute(attributename)
            if var.is_multi():
                # gradual only for multi
                # to preserve order on lists used in attributes
                # while allowing gradual execution on relations
                reqs = valueexpression.requires_emit_gradual(
                    self_resolver, queue, GradualSetAttributeHelper(self, object_instance, attributename, var)
                )
            else:
                reqs = valueexpression.requires_emit(self_resolver, queue)
            SetAttributeHelper(queue, self_resolver, var, reqs, valueexpression, self, object_instance, attributename)

        # generate an implementation
        for stmt in type_class.get_sub_constructor():
            stmt.emit(object_instance, queue)

        object_instance.trackers.append(queue.get_tracker())

        return object_instance

    def add_attribute(self, lname: LocatableString, value: ExpressionStatement) -> None:
        """
        Add an attribute to this constructor call
        """
        name = str(lname)
        if name not in self.__attributes:
            self.__attributes[name] = value
            self.__attribute_locations[name] = lname
            self.anchors.append(AttributeReferenceAnchor(lname.get_location(), lname.namespace, self.class_type, name))
            self.anchors.extend(value.get_anchors())
        else:
            raise RuntimeException(
                self, "The attribute %s in the constructor call of %s is already set." % (name, self.class_type)
            )

    def get_attributes(self) -> Dict[str, ExpressionStatement]:
        """
        Get the attribtues that are set for this constructor call
        """
        return self.__attributes

    def get_wrapped_kwargs(self) -> List["WrappedKwargs"]:
        """
        Get the wrapped kwargs that are set for this constructor call
        """
        return self.__wrapped_kwarg_attributes

    attributes = property(get_attributes)
    wrapped_kwargs = property(get_wrapped_kwargs)

    def _register_dataflow_node(self, graph: DataflowGraph) -> dataflow.InstanceNodeReference:
        """
        Registers the dataflow node for this constructor to the graph if it does not exist yet.
        Returns the node.
        """

        def get_new_node() -> dataflow.InstanceNode:
            assert self.type is not None
            return dataflow.InstanceNode(self.type.get_all_attribute_names())

        assert self.type is not None
        return graph.own_instance_node_for_responsible(self.type, self, get_new_node).reference()

    def get_dataflow_node(self, graph: DataflowGraph) -> dataflow.InstanceNodeReference:
        return self._register_dataflow_node(graph)

    def __repr__(self) -> str:
        """
        The representation of the this statement
        """
        return "Construct(%s)" % (self.class_type)


class WrappedKwargs(ExpressionStatement):
    """
    Keyword arguments wrapped in a dictionary.
    Separate AST node for the type check it provides in the execute method.
    """

    __slots__ = ("dictionary",)

    def __init__(self, dictionary: ExpressionStatement) -> None:
        super().__init__()
        self.dictionary: ExpressionStatement = dictionary

    def __repr__(self) -> str:
        return "**%s" % repr(self.dictionary)

    def normalize(self, *, lhs_attribute: Optional[AttributeAssignmentLHS] = None) -> None:
        self.dictionary.normalize()

    def get_all_eager_promises(self) -> Iterator["StaticEagerPromise"]:
        return chain(super().get_all_eager_promises(), self.dictionary.get_all_eager_promises())

    def requires(self) -> List[str]:
        return self.dictionary.requires()

    def requires_emit(self, resolver: Resolver, queue: QueueScheduler) -> Dict[object, VariableABC[object]]:
        requires: Dict[object, VariableABC[object]] = super().requires_emit(resolver, queue)
        requires.update(self.dictionary.requires_emit(resolver, queue))
        return requires

    def execute(self, requires: Dict[object, object], resolver: Resolver, queue: QueueScheduler) -> List[Tuple[str, object]]:
        super().execute(requires, resolver, queue)
        dct: object = self.dictionary.execute(requires, resolver, queue)
        if not isinstance(dct, Dict):
            raise TypingException(self, "The ** operator can only be applied to dictionaries")
        return list(dct.items())


class IndexCollisionException(RuntimeException):
    """Exception raised when an index collision is detected"""

    def __init__(
        self,
        msg: str,
        collisions: abc.Mapping[tuple[str, ...], Instance],
        constructor: Constructor,
    ) -> None:
        super().__init__(stmt=constructor, msg=msg)
        self.collisions: abc.Mapping[tuple[str, ...], Instance] = collisions
        self.constructor: Constructor = constructor

    def importantance(self) -> int:
        return 10



"""
    Copyright 2019 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import asyncio
import concurrent
import logging
import uuid

import pytest

from inmanta import config, data, protocol
from inmanta.agent import Agent, reporting
from inmanta.agent.handler import HandlerContext, InvalidOperation
from inmanta.data.model import AttributeStateChange
from inmanta.resources import Id, PurgeableResource
from inmanta.server import SLICE_AGENT_MANAGER, SLICE_SESSION_MANAGER
from inmanta.server.bootloader import InmantaBootloader
from utils import retry_limited

logger = logging.getLogger(__name__)


@pytest.mark.slowtest
async def test_agent_get_status(server, environment, agent):
    clients = server.get_slice(SLICE_SESSION_MANAGER)._sessions.values()
    assert len(clients) == 1
    clients = [x for x in clients]
    client = clients[0].get_client()
    status = await client.get_status()
    status = status.get_result()
    for name in reporting.reports.keys():
        assert name in status and status[name] != "ERROR"
    assert status.get("env") is None


def test_context_changes():
    """Test registering changes in the handler context"""
    resource = PurgeableResource(Id.parse_id("std::File[agent,path=/test],v=1"))
    ctx = HandlerContext(resource)

    # use attribute change attributes
    ctx.update_changes({"value": AttributeStateChange(current="a", desired="b")})
    assert len(ctx.changes) == 1

    # use dict
    ctx.update_changes({"value": dict(current="a", desired="b")})
    assert len(ctx.changes) == 1
    assert isinstance(ctx.changes["value"], AttributeStateChange)

    # use dict with empty string
    ctx.update_changes({"value": dict(current="", desired="value")})
    assert len(ctx.changes) == 1
    assert ctx.changes["value"].current == ""
    assert ctx.changes["value"].desired == "value"

    # use tuple
    ctx.update_changes({"value": ("a", "b")})
    assert len(ctx.changes) == 1
    assert isinstance(ctx.changes["value"], AttributeStateChange)

    # use wrong arguments
    with pytest.raises(InvalidOperation):
        ctx.update_changes({"value": ("a", "b", 3)})

    with pytest.raises(InvalidOperation):
        ctx.update_changes({"value": ["a", "b"]})

    with pytest.raises(InvalidOperation):
        ctx.update_changes({"value": "test"})


@pytest.fixture(scope="function")
async def async_started_agent(server_config):
    """
    Start agent with the use_autostart_agent_map option to true.
    agent.start() is executed in the background, since connection to the server will fail.
    """
    config.Config.set("config", "use_autostart_agent_map", "true")

    env_id = uuid.uuid4()
    a = Agent(hostname="node1", environment=env_id, agent_map=None, code_loader=False)
    task = asyncio.ensure_future(a.start())
    yield a
    task.cancel()
    while not task.done():
        # The CancelledError is only thrown on the next invocation of the event loop.
        # Wait until the cancellation has finished.
        await asyncio.sleep(0)


@pytest.fixture(scope="function")
async def startable_server(server_config):
    """
    This fixture returns the bootloader of a server which is not yet started.
    """
    bootloader = InmantaBootloader()
    yield bootloader
    try:
        await bootloader.stop(timeout=15)
    except concurrent.futures.TimeoutError:
        logger.exception("Timeout during stop of the server in teardown")


async def test_agent_cannot_retrieve_autostart_agent_map(async_started_agent, startable_server, caplog):
    """
    When an agent with the config option use_autostart_agent_map set to true, cannot retrieve the autostart_agent_map
    from the server at startup, the process should retry. This test verifies that the retry happens correctly.
    """
    client = protocol.Client("client")

    def retry_occured() -> bool:
        return caplog.text.count("Failed to retrieve the autostart_agent_map setting from the server.") > 2

    # Agent cannot contact server, since server is not started yet.
    await retry_limited(retry_occured, 10)

    # Start server
    await startable_server.start()

    # Create project
    result = await client.create_project("env-test")
    assert result.code == 200
    project_id = result.result["project"]["id"]

    # Create environment
    result = await client.create_environment(project_id=project_id, name="dev", environment_id=async_started_agent.environment)
    assert result.code == 200

    # set agent in agent map
    result = await client.get_setting(tid=async_started_agent.environment, id=data.AUTOSTART_AGENT_MAP)
    assert result.code == 200
    result = await client.set_setting(
        tid=async_started_agent.environment,
        id=data.AUTOSTART_AGENT_MAP,
        value={"agent1": "localhost"} | result.result["value"],
    )
    assert result.code == 200

    # Assert agent managed to establish session with the server
    agent_manager = startable_server.restserver.get_slice(SLICE_AGENT_MANAGER)
    await retry_limited(lambda: (async_started_agent.environment, "agent1") in agent_manager.tid_endpoint_to_session, 10)


async def test_set_agent_map(server, environment, agent_factory):
    """
    This test verifies whether an agentmap is set correct when set via:
        1) The constructor
        2) Via the agent-map configuration option
    """
    env_id = uuid.UUID(environment)
    agent_map = {"agent1": "localhost"}

    # Set agentmap in constructor
    agent1 = await agent_factory(hostname="node1", environment=env_id, agent_map=agent_map)
    assert agent1.agent_map == agent_map

    # Set agentmap via config option
    config.Config.set("config", "agent-map", "agent2=localhost")
    agent2 = await agent_factory(hostname="node2", environment=env_id)
    assert agent2.agent_map == {"agent2": "localhost"}

    # When both are set, the constructor takes precedence
    config.Config.set("config", "agent-map", "agent3=localhost")
    agent3 = await agent_factory(hostname="node3", environment=env_id, agent_map=agent_map)
    assert agent3.agent_map == agent_map


async def test_hostname(server, environment, agent_factory):
    """
    This test verifies whether the hostname of an agent is set correct when set via:
        1) The constructor
        2) Via the agent-names configuration option
    """
    env_id = uuid.UUID(environment)

    # Set hostname in constructor
    agent1 = await agent_factory(hostname="node1", environment=env_id)
    assert list(agent1.get_end_point_names()) == ["node1"]

    # Set hostname via config option
    config.Config.set("config", "agent-names", "test123,test456")
    agent2 = await agent_factory(environment=env_id)
    assert sorted(list(agent2.get_end_point_names())) == sorted(["test123", "test456"])

    # When both are set, the constructor takes precedence
    agent3 = await agent_factory(hostname="node3", environment=env_id)
    assert list(agent3.get_end_point_names()) == ["node3"]


async def test_update_agent_map(server, environment, agent_factory):
    """
    If the URI of an enabled agent changes, it should still be enabled after the change
    """
    env_id = uuid.UUID(environment)
    agent_map = {"node1": "localhost"}

    agent1 = await agent_factory(hostname="node1", environment=env_id, agent_map=agent_map)
    assert agent1.agent_map == agent_map

    agent1.unpause("node1")

    await agent1._update_agent_map({"node1": "localhost2"})

    assert agent1._instances["node1"].is_enabled()



"""
    Copyright 2023 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import logging
import os
import sys
from argparse import Namespace
from typing import Optional, TextIO

import colorlog
from colorlog.formatter import LogColors

from inmanta import const
from inmanta.stable_api import stable_api


def _is_on_tty() -> bool:
    return (hasattr(sys.stdout, "isatty") and sys.stdout.isatty()) or const.ENVIRON_FORCE_TTY in os.environ


"""
This dictionary maps the Inmanta log levels to the corresponding Python log levels
"""
log_levels = {
    "0": logging.ERROR,
    "1": logging.WARNING,
    "2": logging.INFO,
    "3": logging.DEBUG,
    "4": 2,
    "ERROR": logging.ERROR,
    "WARNING": logging.WARNING,
    "INFO": logging.INFO,
    "DEBUG": logging.DEBUG,
    "TRACE": 2,
}


class Options(Namespace):
    """
    The Options class provides a way to configure the InmantaLoggerConfig with the following attributes:
    - `log_file`: if this attribute is set, the logs will be written to the specified file instead of the stream
      specified in `get_instance`.
    - `log_file_level`: the Inmanta logging level for the file handler (if `log_file` is set).
        The possible inmanta log levels and their associated python log level are defined in the
        inmanta.logging.log_levels dictionary.
    - `verbose`: the verbosity level of the log messages. can be a number from 0 to 4.
        if a bigger number is provided, 4 will be used. Refer to log_file_level for the explanation of each level.
        default is 1 (WARNING)
    - `timed`: if true,  adds the time to the formatter in the log lines.
    """

    log_file: Optional[str]
    log_file_level: str = "INFO"
    verbose: int = 1
    timed: bool = False


@stable_api
class InmantaLoggerConfig:
    """
    A class that provides logging functionality for Inmanta projects.

    Usage:
    To use this class, you first need to call the `get_instance`. This method takes a `stream` argument
    that specifies where the log messages should be sent to. If no `stream` is provided,
    the log messages will be sent to standard output.

    You can then call the `apply_options` method to configure the logging options.

    The setup is not done in one step as we want logs for the cmd_parser, which will provide the options needed to configure
    the 'final' logger with `apply_options`.

    for more fine-grained configuration the following functions can be used as well:
        - `set_log_level`
        - `set_log_formatter`
        - `set_logfile_location`
    """

    _instance: Optional["InmantaLoggerConfig"] = None

    def __init__(self, stream: TextIO = sys.stdout) -> None:
        """
        Set up the logging handler for Inmanta

        :param stream: The stream to send log messages to. Default is standard output (sys.stdout).
        """
        self._options_applied = False
        self._handler: logging.Handler = logging.StreamHandler(stream=stream)
        self.set_log_level("INFO")
        formatter = self._get_log_formatter_for_stream_handler(timed=False)
        self.set_log_formatter(formatter)

        logging.root.handlers = []
        logging.root.addHandler(self._handler)
        logging.root.setLevel(0)

    @classmethod
    @stable_api
    def get_instance(cls, stream: TextIO = sys.stdout) -> "InmantaLoggerConfig":
        """
        This method should be used to obtain an instance of this class, because this class is a singleton.

        :param stream: The stream to send log messages to. Default is standard output (sys.stdout)
        """
        if cls._instance:
            if not isinstance(cls._instance._handler, logging.StreamHandler):
                raise Exception("Instance already exists with a different handler")
            elif isinstance(cls._instance._handler, logging.StreamHandler) and cls._instance._handler.stream != stream:
                raise Exception("Instance already exists with a different stream")
        else:
            cls._instance = cls(stream)
        return cls._instance

    @classmethod
    @stable_api
    def clean_instance(cls) -> None:
        """
        This method should be used to clean up an instance of this class

        """
        if cls._instance and cls._instance._handler:
            cls._instance._handler.close()
            logging.root.removeHandler(cls._instance._handler)
        cls._instance = None

    @stable_api
    def apply_options(self, options: Options) -> None:
        """
        Apply the logging options to the current handler. A handler should have been created before

        :param options: The Option object coming from the command line. This function uses the following
            attributes: log_file, log_file_level, verbose, timed
        """
        if self._options_applied:
            raise Exception("Options can only be applied once to a handler.")
        self._options_applied = True
        if options.log_file:
            self.set_logfile_location(options.log_file)
            formatter = logging.Formatter(fmt="%(asctime)s %(levelname)-8s %(name)-10s %(message)s")
            self.set_log_formatter(formatter)
            self.set_log_level(options.log_file_level, cli=False)
        else:
            if options.timed:
                formatter = self._get_log_formatter_for_stream_handler(timed=True)
                self.set_log_formatter(formatter)
            self.set_log_level(str(options.verbose))

    @stable_api
    def set_log_level(self, inmanta_log_level: str, cli: bool = True) -> None:
        """
        Set the logging level. A handler should have been created before.
        The possible inmanta log levels and their associated python log level
        are defined in the inmanta.logging.log_levels dictionary.

        :param inmanta_log_level: The inmanta logging level
        :param cli: True if the logs will be outputted to the CLI.
        """
        # maximum of 4 v's
        if inmanta_log_level.isdigit() and int(inmanta_log_level) > 4:
            inmanta_log_level = "4"

        # The minimal log level on the CLI is always WARNING
        if cli and (inmanta_log_level == "ERROR" or (inmanta_log_level.isdigit() and int(inmanta_log_level) < 1)):
            inmanta_log_level = "WARNING"

        # Converts the Inmanta log level to the Python log level
        python_log_level = log_levels[inmanta_log_level]
        self._handler.setLevel(python_log_level)

    @stable_api
    def set_log_formatter(self, formatter: logging.Formatter) -> None:
        """
        Set the log formatter. A handler should have been created before

        :param formatter: The log formatter.
        """
        self._handler.setFormatter(formatter)

    @stable_api
    def set_logfile_location(self, location: str) -> None:
        """
        Set the location of the log file. Be careful that this function will replace the current handler with a new one
        This means that configurations done on the previous handler will be lost.

        :param location: The location of the log file.
        """
        file_handler = logging.handlers.WatchedFileHandler(filename=location, mode="a+")
        if self._handler:
            self._handler.close()
            logging.root.removeHandler(self._handler)
        self._handler = file_handler
        logging.root.addHandler(self._handler)

    @stable_api
    def get_handler(self) -> logging.Handler:
        """
        Get the logging handler

        :return: The logging handler
        """
        return self._handler

    def _get_log_formatter_for_stream_handler(self, timed: bool) -> logging.Formatter:
        log_format = "%(asctime)s " if timed else ""
        if _is_on_tty():
            log_format += "%(log_color)s%(name)-25s%(levelname)-8s%(reset)s%(blue)s%(message)s"
            formatter = MultiLineFormatter(
                log_format,
                reset=True,
                log_colors={"DEBUG": "cyan", "INFO": "green", "WARNING": "yellow", "ERROR": "red", "CRITICAL": "red"},
            )
        else:
            log_format += "%(name)-25s%(levelname)-8s%(message)s"
            formatter = MultiLineFormatter(
                log_format,
                reset=False,
                no_color=True,
            )
        return formatter


class MultiLineFormatter(colorlog.ColoredFormatter):
    """
    Formatter for multi-line log records.

    This class extends the `colorlog.ColoredFormatter` class to provide a custom formatting method for log records that
    span multiple lines.
    """

    def __init__(
        self,
        fmt: Optional[str] = None,
        *,
        # keep interface minimal: only include fields we actually use
        log_colors: Optional[LogColors] = None,
        reset: bool = True,
        no_color: bool = False,
    ):
        """
        Initialize a new `MultiLineFormatter` instance.

        :param fmt: Optional string specifying the log record format.
        :param log_colors: Optional `LogColors` object mapping log level names to color codes.
        :param reset: Boolean indicating whether to reset terminal colors at the end of each log record.
        :param no_color: Boolean indicating whether to disable colors in the output.
        """
        super().__init__(fmt, log_colors=log_colors, reset=reset, no_color=no_color)
        self.fmt = fmt

    def get_header_length(self, record: logging.LogRecord) -> int:
        """
        Get the header length of a given log record.

        :param record: The `logging.LogRecord` object for which to calculate the header length.
        :return: The length of the header in the log record, without color codes.
        """
        # to get the length of the header we want to get the header without the color codes
        formatter = colorlog.ColoredFormatter(
            fmt=self.fmt,
            log_colors=self.log_colors,
            reset=False,
            no_color=True,
        )
        header = formatter.format(
            logging.LogRecord(
                name=record.name,
                level=record.levelno,
                pathname=record.pathname,
                lineno=record.lineno,
                msg="",
                args=(),
                exc_info=None,
            )
        )
        return len(header)

    def format(self, record: logging.LogRecord) -> str:
        """
        Format a log record with added indentation.

        :param record: The `logging.LogRecord` object to format.
        :return: The formatted log record as a string.
        """
        indent: str = " " * self.get_header_length(record)
        head, *tail = super().format(record).splitlines(True)
        return head + "".join(indent + line for line in tail)



"""
    Copyright 2020 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
from asyncpg import Connection

DISABLED = False


async def update(connection: Connection) -> None:
    await connection.execute(
        """
        CREATE INDEX resourceaction_environment_version_started_index ON resourceaction(environment,version,started DESC);
        """
    )
    await enforce_unique_agent_instances(connection)


async def enforce_unique_agent_instances(connection: Connection) -> None:
    """
    Deletes duplicate AgentInstance records and adds a uniqueness constraint.
    """
    async with connection.transaction():
        await connection.execute(
            """
            DELETE FROM public.agentinstance a
            USING public.agentinstance b
            WHERE a.id < b.id AND a.tid = b.tid AND a.process = b.process AND a.name = b.name
            ;

            ALTER TABLE public.agentinstance ADD CONSTRAINT agentinstance_unique UNIQUE (tid, process, name);
            """
        )



"""
    Copyright 2019 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import datetime
import uuid
from enum import Enum
from itertools import chain
from typing import Any, ClassVar, Dict, List, NewType, Optional, Union

import pydantic
import pydantic.schema
from pydantic import Extra, root_validator, validator
from pydantic.fields import ModelField

import inmanta
import inmanta.ast.export as ast_export
from inmanta import const, data, protocol, resources
from inmanta.stable_api import stable_api
from inmanta.types import ArgumentTypes, JsonType, SimpleTypes, StrictNonIntBool

# This reference to the actual pydantic field_type_schema method is only loaded once
old_field_type_schema = pydantic.schema.field_type_schema


def patch_pydantic_field_type_schema() -> None:
    """
    This ugly patch fixes the serialization of models containing Optional in them.
    https://github.com/samuelcolvin/pydantic/issues/1270

    The fix for this issue will be included in pydantic V2.
    """

    def patch_nullable(field: ModelField, **kwargs):
        f_schema, definitions, nested_models = old_field_type_schema(field, **kwargs)
        if field.allow_none:
            f_schema["nullable"] = True
        return f_schema, definitions, nested_models

    pydantic.schema.field_type_schema = patch_nullable


def api_boundary_datetime_normalizer(value: datetime.datetime) -> datetime.datetime:
    if value.tzinfo is None:
        return value.replace(tzinfo=datetime.timezone.utc)
    else:
        return value


def validator_timezone_aware_timestamps(value: object) -> object:
    """
    A Pydantic validator to ensure that all datetime times are timezone aware.
    """
    if isinstance(value, datetime.datetime):
        return api_boundary_datetime_normalizer(value)
    else:
        return value


@stable_api
class BaseModel(pydantic.BaseModel):
    """
    Base class for all data objects in Inmanta
    """

    _normalize_timestamps: ClassVar[classmethod] = pydantic.validator("*", allow_reuse=True)(
        validator_timezone_aware_timestamps
    )

    class Config:
        """
        Pydantic config.
        """

        # Populate models with the value property of enums, rather than the raw enum.
        # This is useful to serialise model.dict() later
        use_enum_values = True


class ExtensionStatus(BaseModel):
    """
    Status response for extensions loaded in the server
    """

    name: str
    version: str
    package: str


class SliceStatus(BaseModel):
    """
    Status response for slices loaded in the the server
    """

    name: str
    status: Dict[str, ArgumentTypes]


class FeatureStatus(BaseModel):
    """
    Status of the feature
    """

    slice: str
    name: str
    value: Optional[Any]


class StatusResponse(BaseModel):
    """
    Response for the status method call
    """

    product: str
    edition: str
    version: str
    license: Union[str, Dict[str, SimpleTypes]]
    extensions: List[ExtensionStatus]
    slices: List[SliceStatus]
    features: List[FeatureStatus]


@stable_api
class CompileData(BaseModel):
    """
    Top level structure of compiler data to be exported.
    """

    errors: List[ast_export.Error]
    """
        All errors occurred while trying to compile.
    """


class CompileRunBase(BaseModel):
    id: uuid.UUID
    remote_id: Optional[uuid.UUID]
    environment: uuid.UUID
    requested: Optional[datetime.datetime]
    started: Optional[datetime.datetime]

    do_export: bool
    force_update: bool
    metadata: JsonType
    environment_variables: Dict[str, str]

    partial: bool
    removed_resource_sets: list[str]

    exporter_plugin: Optional[str]

    notify_failed_compile: Optional[bool]
    failed_compile_message: Optional[str]


class CompileRun(CompileRunBase):
    compile_data: Optional[CompileData]


class CompileReport(CompileRunBase):
    completed: Optional[datetime.datetime]
    success: Optional[bool]
    version: Optional[int]


class CompileRunReport(BaseModel):
    id: uuid.UUID
    started: datetime.datetime
    completed: Optional[datetime.datetime]
    command: str
    name: str
    errstream: str
    outstream: str
    returncode: Optional[int]


class CompileDetails(CompileReport):
    compile_data: Optional[CompileData]
    reports: Optional[List[CompileRunReport]]


ResourceVersionIdStr = NewType("ResourceVersionIdStr", str)  # Part of the stable API
"""
    The resource id with the version included.
"""

ResourceIdStr = NewType("ResourceIdStr", str)  # Part of the stable API
"""
    The resource id without the version
"""

ResourceType = NewType("ResourceType", str)
"""
    The type of the resource
"""


class AttributeStateChange(BaseModel):
    """
    Changes in the attribute
    """

    current: Optional[Any] = None
    desired: Optional[Any] = None

    @validator("current", "desired")
    @classmethod
    def check_serializable(cls, v: Optional[Any]) -> Optional[Any]:
        """
        Verify whether the value is serializable (https://github.com/inmanta/inmanta-core/issues/3470)
        """
        try:
            protocol.common.json_encode(v)
        except TypeError:
            if inmanta.RUNNING_TESTS:
                # Fail the test when the value is not serializable
                raise Exception(f"Failed to serialize attribute {v}")
            else:
                # In production, try to cast the non-serializable value to str to prevent the handler from failing.
                return str(v)
        return v


EnvSettingType = Union[StrictNonIntBool, int, float, str, Dict[str, Union[str, int, StrictNonIntBool]]]


class Environment(BaseModel):
    """
    An inmanta environment.

    :note: repo_url and repo_branch will be moved to the settings.
    """

    id: uuid.UUID
    name: str
    project_id: uuid.UUID
    repo_url: str
    repo_branch: str
    settings: Dict[str, EnvSettingType]
    halted: bool
    description: Optional[str]
    icon: Optional[str]


class Project(BaseModel):
    """
    An inmanta environment.
    """

    id: uuid.UUID
    name: str
    environments: List[Environment]


class EnvironmentSetting(BaseModel):
    """A class to define a new environment setting.

    :param name: The name of the setting.
    :param type: The type of the value. This type is mainly used for documentation purpose.
    :param default: An optional default value for this setting. When a default is set and the
                    is requested from the database, it will return the default value and also store
                    the default value in the database.
    :param doc: The documentation/help string for this setting
    :param recompile: Trigger a recompile of the model when a setting is updated?
    :param update_model: Update the configuration model (git pull on project and repos)
    :param agent_restart: Restart autostarted agents when this settings is updated.
    :param allowed_values: list of possible values (if type is enum)
    """

    name: str
    type: str
    default: EnvSettingType
    doc: str
    recompile: bool
    update_model: bool
    agent_restart: bool
    allowed_values: Optional[List[EnvSettingType]]


class EnvironmentSettingsReponse(BaseModel):
    settings: Dict[str, EnvSettingType]
    definition: Dict[str, EnvironmentSetting]


class ModelMetadata(BaseModel):
    """Model metadata"""

    inmanta_compile_state: const.Compilestate = const.Compilestate.success
    message: str
    type: str
    extra_data: Optional[JsonType]

    class Config:
        fields = {"inmanta_compile_state": {"alias": "inmanta:compile:state"}}


class ModelVersionInfo(BaseModel):
    """Version information that can be associated with an orchestration model

    :param export_metadata: Metadata associated with this version
    :param model: A serialization of the complete orchestration model
    """

    export_metadata: ModelMetadata
    model: Optional[JsonType]


class ResourceMinimal(BaseModel):
    """
    Represents a resource object as it comes in over the API. Provides strictly required validation only.
    """

    id: ResourceVersionIdStr

    @classmethod
    @validator("id")
    def id_is_resource_version_id(cls, v):
        if resources.Id.is_resource_version_id(v):
            return v
        raise ValueError(f"id {v} is not of type ResourceVersionIdStr")

    class Config:
        extra = Extra.allow


class Resource(BaseModel):
    environment: uuid.UUID
    model: int
    resource_id: ResourceIdStr
    resource_type: ResourceType
    resource_version_id: ResourceVersionIdStr
    resource_id_value: str
    agent: str
    last_deploy: Optional[datetime.datetime]
    attributes: JsonType
    status: const.ResourceState
    resource_set: Optional[str]


class ResourceAction(BaseModel):
    environment: uuid.UUID
    version: int
    resource_version_ids: List[ResourceVersionIdStr]
    action_id: uuid.UUID
    action: const.ResourceAction
    started: datetime.datetime
    finished: Optional[datetime.datetime]
    messages: Optional[List[JsonType]]
    status: Optional[const.ResourceState]
    changes: Optional[JsonType]
    change: Optional[const.Change]
    send_event: Optional[bool] = None  # Deprecated field


class ResourceDeploySummary(BaseModel):
    """
    :param total: The total number of resources
    :param by_state: The number of resources by state in the latest released version
    """

    total: int
    by_state: Dict[str, int]

    @classmethod
    def create_from_db_result(cls, summary_by_state: Dict[str, int]) -> "ResourceDeploySummary":
        full_summary_by_state = cls._ensure_summary_has_all_states(summary_by_state)
        total = cls._count_all_resources(full_summary_by_state)
        return ResourceDeploySummary(by_state=full_summary_by_state, total=total)

    @classmethod
    def _ensure_summary_has_all_states(cls, summary_by_state: Dict[str, int]) -> Dict[str, int]:
        full_summary = summary_by_state.copy()
        for state in const.ResourceState:
            if state not in summary_by_state.keys() and state != const.ResourceState.dry:
                full_summary[state] = 0
        return full_summary

    @classmethod
    def _count_all_resources(cls, summary_by_state: Dict[str, int]) -> int:
        return sum(resource_count for resource_count in summary_by_state.values())


class LogLine(BaseModel):
    class Config:
        """
        Pydantic config.
        """

        # Override the setting from the BaseModel class as such that the level field is
        # serialises using the name of the enum instead of its value. This is required
        # to make sure that data sent to the API endpoints resource_action_update
        # and resource_deploy_done are serialized consistently using the name of the enum.
        use_enum_values = False

    level: const.LogLevel
    msg: str
    args: List[Optional[ArgumentTypes]] = []
    kwargs: JsonType = {}
    timestamp: datetime.datetime


class ResourceIdDetails(BaseModel):
    resource_type: ResourceType
    agent: str
    attribute: str
    resource_id_value: str


class OrphanedResource(str, Enum):
    orphaned = "orphaned"


class StrEnum(str, Enum):
    """Enum where members are also (and must be) strs"""


ReleasedResourceState = StrEnum(
    "ReleasedResourceState", [(i.name, i.value) for i in chain(const.ResourceState, OrphanedResource)]
)


class VersionedResource(BaseModel):
    resource_id: ResourceIdStr
    resource_version_id: ResourceVersionIdStr
    id_details: ResourceIdDetails
    requires: List[ResourceVersionIdStr]

    @property
    def all_fields(self) -> Dict[str, Any]:
        return {**self.dict(), **self.id_details.dict()}


class LatestReleasedResource(VersionedResource):
    status: ReleasedResourceState


class PagingBoundaries:
    """Represents the lower and upper bounds that should be used for the next and previous pages
    when listing domain entities"""

    def __init__(
        self,
        start: Optional["inmanta.data.PRIMITIVE_SQL_TYPES"],  # Can be none if user selected field is nullable
        end: Optional["inmanta.data.PRIMITIVE_SQL_TYPES"],  # Can be none if user selected field is nullable
        first_id: Optional["inmanta.data.PRIMITIVE_SQL_TYPES"],  # Can be none if single keyed
        last_id: Optional["inmanta.data.PRIMITIVE_SQL_TYPES"],  # Can be none if single keyed
    ) -> None:
        self.start = start
        self.end = end
        self.first_id = first_id
        self.last_id = last_id


class ResourceDetails(BaseModel):
    """The details of a resource
    :param resource_id: The id of the resource
    :param resource_type: The type of the resource
    :param agent: The agent associated with this resource
    :param id_attribute: The name of the identifying attribute of the resource
    :param id_attribute_value: The value of the identifying attribute of the resource
    :param attributes: The attributes of the resource
    """

    resource_id: ResourceIdStr
    resource_type: ResourceType
    agent: str
    id_attribute: str
    id_attribute_value: str
    attributes: JsonType


class VersionedResourceDetails(ResourceDetails):
    """The details of a resource version
    :param resource_version_id: The id of the resource
    :param version: The version of the resource
    """

    resource_version_id: ResourceVersionIdStr
    version: int

    @root_validator
    @classmethod
    def ensure_version_field_set_in_attributes(cls, v: JsonType) -> JsonType:
        # Due to a bug, the version field has always been present in the attributes dictionary.
        # This bug has been fixed in the database. For backwards compatibility reason we here make sure that the
        # version field is present in the attributes dictionary served out via the API.
        if "version" not in v["attributes"]:
            v["attributes"]["version"] = v["version"]
        return v


class ReleasedResourceDetails(ResourceDetails):
    """The details of a released resource
    :param last_deploy: The value of the last_deploy on the latest released version of the resource
    :param first_generated_time: The first time this resource was generated
    :param first_generated_version: The first model version this resource was in
    :param status: The current status of the resource
    :param requires_status: The id and status of the resources this resource requires
    """

    last_deploy: Optional[datetime.datetime]
    first_generated_time: datetime.datetime
    first_generated_version: int
    status: ReleasedResourceState
    requires_status: Dict[ResourceIdStr, ReleasedResourceState]


class ResourceHistory(BaseModel):
    resource_id: ResourceIdStr
    date: datetime.datetime
    attributes: JsonType
    attribute_hash: str
    requires: List[ResourceIdStr]


class ResourceLog(LogLine):
    action_id: uuid.UUID
    action: const.ResourceAction


class ResourceDiffStatus(str, Enum):
    added = "added"
    modified = "modified"
    deleted = "deleted"
    unmodified = "unmodified"
    agent_down = "agent_down"
    undefined = "undefined"
    skipped_for_undefined = "skipped_for_undefined"


class AttributeDiff(BaseModel):
    """
    :param from_value: The value of the attribute in the earlier version
    :param to_value: The value of the attribute in the later version
    :param from_value_compare: A stringified, diff-friendly form of the 'from_value' field
    :param to_value_compare: A stringified, diff-friendly form of the 'to_value' field
    """

    from_value: Optional[object] = None
    to_value: Optional[object] = None
    from_value_compare: str
    to_value_compare: str


class ResourceDiff(BaseModel):
    """
    :param resource_id: The id of the resource the diff is about (without version)
    :param attributes: The diff between the attributes of two versions of the resource
    :param status: The kind of diff between the versions of the resource
    """

    resource_id: ResourceIdStr
    attributes: Dict[str, AttributeDiff]
    status: ResourceDiffStatus


class Parameter(BaseModel):
    id: uuid.UUID
    name: str
    value: str
    environment: uuid.UUID
    source: str
    updated: Optional[datetime.datetime]
    metadata: Optional[JsonType]


class Fact(Parameter):
    resource_id: ResourceIdStr


class Agent(BaseModel):
    """
    :param environment: Id of the agent's environment
    :param name: The name of the agent
    :param last_failover: The time of the last failover
    :param paused: Whether the agent is paused or not
    :param unpause_on_resume: Whether the agent should be unpaused when the environment is resumed
    :param status: The current status of the agent
    :param process_id: The id of the agent process that belongs to this agent, if there is one
    :param process_name: The name of the agent process that belongs to this agent, if there is one
    """

    environment: uuid.UUID
    name: str
    last_failover: Optional[datetime.datetime]
    paused: bool
    process_id: Optional[uuid.UUID]
    process_name: Optional[str]
    unpause_on_resume: Optional[bool]
    status: const.AgentStatus


class AgentProcess(BaseModel):
    sid: uuid.UUID
    hostname: str
    environment: uuid.UUID
    first_seen: Optional[datetime.datetime]
    last_seen: Optional[datetime.datetime]
    expired: Optional[datetime.datetime]
    state: Optional[Dict[str, Union[Dict[str, List[str]], Dict[str, str], Dict[str, float], str]]]


class DesiredStateLabel(BaseModel):
    name: str
    message: str


class DesiredStateVersion(BaseModel):
    version: int
    date: datetime.datetime
    total: int
    labels: List[DesiredStateLabel]
    status: const.DesiredStateVersionStatus


class NoPushTriggerMethod(str, Enum):
    no_push = "no_push"


PromoteTriggerMethod = StrEnum(
    "PromoteTriggerMethod", [(i.name, i.value) for i in chain(const.AgentTriggerMethod, NoPushTriggerMethod)]
)


class DryRun(BaseModel):
    id: uuid.UUID
    environment: uuid.UUID
    model: int
    date: Optional[datetime.datetime]
    total: int = 0
    todo: int = 0


class DryRunReport(BaseModel):
    summary: DryRun
    diff: List[ResourceDiff]


class Notification(BaseModel):
    """
    :param id: The id of this notification
    :param environment: The environment this notification belongs to
    :param created: The date the notification was created at
    :param title: The title of the notification
    :param message: The actual text of the notification
    :param severity: The severity of the notification
    :param uri: A link to an api endpoint of the server, that is relevant to the message,
                and can be used to get further information about the problem.
                For example a compile related problem should have the uri: `/api/v2/compilereport/<compile_id>`
    :param read: Whether the notification was read or not
    :param cleared: Whether the notification was cleared or not
    """

    id: uuid.UUID
    environment: uuid.UUID
    created: datetime.datetime
    title: str
    message: str
    severity: const.NotificationSeverity
    uri: str
    read: bool
    cleared: bool


class Source(BaseModel):
    """Model for source code"""

    hash: str
    is_byte_code: bool
    module_name: str
    requirements: List[str]


class EnvironmentMetricsResult(BaseModel):
    """
    A container for metrics as returned by the /metrics endpoint.

    :param start: The starting of the requested aggregation interval.
    :param end: The end of the requested aggregation interval.
    :param timestamps: The timestamps that belongs to the aggregated metrics present in the `metrics` dictionary.
    :param metrics: A dictionary that maps the name of a metric to a list of aggregated datapoints. For metrics that are not
                    grouped on a specific property, this list only contains the values of the metrics. For metrics that
                    are grouped by a specific property, this list contains a dictionary where the key is the grouping
                    attribute and the value is the value of the metric. The value is None when no data is available
                    for that specific time window.
    """

    start: datetime.datetime
    end: datetime.datetime
    timestamps: List[datetime.datetime]
    metrics: Dict[str, List[Optional[Union[float, Dict[str, float]]]]]


class AuthMethod(str, Enum):
    database = "database"
    oidc = "oidc"


class User(BaseModel):
    """A user"""

    username: str
    auth_method: AuthMethod


class LoginReturn(BaseModel):
    """
    Login information

    :param token: A token representing the user's authentication session
    :param user: The user object for which the token was created
    """

    token: str
    user: User


class DiscoveredResource(BaseModel):
    """
    :param discovered_resource_id: The name of the resource
    :param values: The actual resource
    """

    discovered_resource_id: ResourceIdStr
    values: JsonType

    @validator("discovered_resource_id")
    @classmethod
    def discovered_resource_id_is_resource_id(cls, v: str) -> Optional[Any]:
        if resources.Id.is_resource_id(v):
            return v
        raise ValueError(f"id {v} is not of type ResourceIdStr")

    def to_dao(self, env: uuid) -> "data.DiscoveredResource":
        return data.DiscoveredResource(
            discovered_resource_id=self.discovered_resource_id,
            values=self.values,
            discovered_at=datetime.datetime.now(),
            environment=env,
        )



"""
    Copyright 2017 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""


"""
This module enables tracking of object construction.

Tracker object are attached to all created Instances in the field trackers. Every Instance can have one or more Trackers.

If the Tracker object is a ModuleTracker, the object is created at module level
If the Tracker object is an ImplementsTracker, the object is created in an Implementation block

"""

from typing import TYPE_CHECKING, List

if TYPE_CHECKING:
    from inmanta.ast.blocks import BasicBlock
    from inmanta.ast.statements.generator import SubConstructor
    from inmanta.execute.runtime import Instance


class Tracker(object):
    def get_next(self) -> "List[Tracker]":
        return []


class ModuleTracker(Tracker):
    def __init__(self, block: "BasicBlock") -> None:
        self.block = block
        self.namespace = block.namespace


class ImplementsTracker(Tracker):
    def __init__(self, subc: "SubConstructor", instance: "Instance") -> None:
        self.instance = instance
        self.subc = subc
        self.implements = subc.implements
        self.implementations = self.implements.implementations

    def get_next(self) -> "List[Tracker]":
        return self.instance.trackers



"""
    Copyright 2021 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import json
from datetime import datetime
from typing import Dict, List

from asyncpg import Connection, Record

from inmanta import const

DISABLED = False


TIMESTAMP_COLUMNS: Dict[str, List[str]] = {
    "agent": ["last_failover"],
    "agentinstance": ["expired"],
    "agentprocess": ["first_seen", "last_seen", "expired"],
    "compile": ["started", "completed", "requested"],
    "configurationmodel": ["date"],
    "dryrun": ["date"],
    "parameter": ["updated"],
    "report": ["started", "completed"],
    "resourceaction": ["started", "finished"],
    "resource": ["last_deploy"],
}


async def update(connection: Connection) -> None:
    # update all timestamp types
    await connection.execute(
        "\n".join(
            f"ALTER TABLE public.{table} %s;"
            % ", ".join(f"ALTER COLUMN {column} TYPE TIMESTAMP WITH TIME ZONE" for column in columns)
            for table, columns in TIMESTAMP_COLUMNS.items()
        )
    )

    # update timestamps embedded in jsonb types
    def transform_message(message: str) -> str:
        obj: Dict = json.loads(message)
        if "timestamp" in obj:
            obj["timestamp"] = (
                datetime.strptime(obj["timestamp"], const.TIME_ISOFMT).astimezone().isoformat(timespec="microseconds")
            )
        return json.dumps(obj)

    records: List[Record] = await connection.fetch("SELECT action_id, messages FROM public.resourceaction")
    await connection.executemany(
        """
        UPDATE public.resourceaction
        SET messages = $1
        WHERE action_id = $2
        """,
        [
            (
                None if record["messages"] is None else [transform_message(msg) for msg in record["messages"]],
                record["action_id"],
            )
            for record in records
        ],
    )



"""
    Copyright 2019 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
from asyncpg import Connection

DISABLED = False


async def update(connection: Connection) -> None:
    schema = """
ALTER TABLE public.environment
    ADD COLUMN last_version integer DEFAULT 0;

UPDATE public.environment AS e SET last_version =
    (SELECT COALESCE(
        (SELECT MAX(version) FROM public.configurationmodel AS c WHERE c.environment=e.id),
        0
    ));

ALTER TABLE public.resource
    ADD COLUMN resource_type varchar;

UPDATE public.resource
SET resource_type=substring(resource_id from '(.*)\\[');

-- Set NOT NULL constraint after column is populated
ALTER TABLE public.resource
    ALTER COLUMN resource_type SET NOT NULL;

CREATE INDEX resource_environment_resource_type_index ON public.resource (environment, resource_type);
"""
    async with connection.transaction():
        await connection.execute(schema)



"""
    Copyright 2021 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import datetime
from abc import ABC, abstractmethod
from typing import Callable, Dict, List, Optional, Tuple, Type, TypeVar

import dateutil
import more_itertools
from pydantic import BaseModel, ValidationError, validator

from inmanta import const
from inmanta.data import DateRangeConstraint, QueryFilter, QueryType, RangeConstraint, RangeOperator
from inmanta.data.model import ReleasedResourceState


class InvalidFilter(Exception):
    def __init__(self, message: str, *args: object) -> None:
        super().__init__(message, *args)
        self.message = message


def parse_single_value(v: object) -> object:
    """
    Transform list values to their single element value.
    """
    if isinstance(v, list):
        return more_itertools.one(
            v,
            too_short=ValueError("Empty filter provided"),
            too_long=ValueError(f"Multiple values provided for filter: {v}"),
        )

    return v


def parse_range_value_to_date(single_constraint: str, value: str) -> datetime.datetime:
    try:
        datetime_obj: datetime.datetime = dateutil.parser.isoparse(value)
    except ValueError:
        raise ValueError("Invalid range constraint %s: '%s' is not a valid datetime" % (single_constraint, value))
    else:
        return datetime_obj if datetime_obj.tzinfo is not None else datetime_obj.replace(tzinfo=datetime.timezone.utc)


def parse_range_value_to_int(single_constraint: str, value: str) -> int:
    try:
        return int(value)
    except ValueError:
        raise ValueError("Invalid range constraint %s: '%s' is not an integer" % (single_constraint, value))


S = TypeVar("S", int, datetime.datetime)


def get_range_operator_parser(
    parse_value_to_type: Callable[[str, str], S]
) -> Callable[[object, object], Optional[List[Tuple[RangeOperator, S]]]]:
    def parse_range_operator(v: object) -> Optional[List[Tuple[RangeOperator, S]]]:
        """
        Transform list of "<lt|le|gt|ge>:<x>" constraint specifiers to typed objects.
        """

        def transform_single(single: str, parse_value_to_type: Callable[[str, str], S]) -> Tuple[RangeOperator, S]:
            split: List[str] = single.split(":", maxsplit=1)
            if len(split) != 2:
                raise ValueError("Invalid range constraint %s, expected '<lt|le|gt|ge>:<x>`" % single)
            operator: RangeOperator
            try:
                operator = RangeOperator.parse(split[0])
            except ValueError:
                raise ValueError(
                    "Invalid range operator %s in constraint %s, expected one of lt, le, gt, ge" % (split[0], single)
                )
            bound = parse_value_to_type(single, split[1])
            return (operator, bound)

        if v is None:
            return None

        if isinstance(v, str):
            return [transform_single(v, parse_value_to_type=parse_value_to_type)]

        if isinstance(v, list) and all(isinstance(x, str) for x in v):
            return [transform_single(x, parse_value_to_type=parse_value_to_type) for x in v]

        raise ValueError(f"value is not a valid list of range constraints: {str(v)}")

    return parse_range_operator


class Filter(ABC, BaseModel):
    """A pydantic model for a specific filter
    Subclasses are expected to have property named `field`, representing the value to be validated.
    The `to_query_type` method describes how to interpret this as filter in for a database query by providing the correct
    `QueryType`
    """

    # Pydantic doesn't support Generic BaseModels on python 3.6

    @abstractmethod
    def to_query_type(self) -> Optional[Tuple[QueryType, object]]:
        """Get the value of the filter with the correct query type"""
        pass


class BooleanEqualityFilter(Filter):
    """Represents a valid boolean which should be handled as an equality filter"""

    field: Optional[bool]
    validate_field: classmethod = validator("field", pre=True, allow_reuse=True)(parse_single_value)

    def to_query_type(self) -> Optional[Tuple[QueryType, object]]:
        if self.field is not None:
            return (QueryType.EQUALS, self.field)
        return None


class BooleanIsNotNullFilter(BooleanEqualityFilter, Filter):
    """Represents a valid boolean which should be handled as an IS_NOT_NULL filter"""

    def to_query_type(self) -> Optional[Tuple[QueryType, object]]:
        if self.field is not None:
            return (QueryType.IS_NOT_NULL, None) if self.field else (QueryType.EQUALS, None)
        return None


class DateRangeFilter(Filter):
    """Represents a valid date range constraint which should be handled as a range filter"""

    field: Optional[DateRangeConstraint]

    @validator("field", pre=True)
    @classmethod
    def parse_requested(cls, v: object) -> Optional[List[Tuple[RangeOperator, datetime.datetime]]]:
        return get_range_operator_parser(parse_range_value_to_date)(v)

    def to_query_type(self) -> Optional[Tuple[QueryType, object]]:
        if self.field:
            return (QueryType.RANGE, self.field)
        return None


class IntRangeFilter(Filter):
    field: Optional[RangeConstraint]

    @validator("field", pre=True)
    @classmethod
    def parse_field(cls, v: object) -> Optional[List[Tuple[RangeOperator, int]]]:
        return get_range_operator_parser(parse_range_value_to_int)(v)

    def to_query_type(self) -> Optional[Tuple[QueryType, object]]:
        if self.field:
            return (QueryType.RANGE, self.field)
        return None


class ContainsPartialFilter(Filter):
    """Represents a valid string list constraint which should be handled as a partial containment filter"""

    field: Optional[List[str]]

    def to_query_type(self) -> Optional[Tuple[QueryType, object]]:
        if self.field:
            return (QueryType.CONTAINS_PARTIAL, self.field)
        return None


class ContainsFilter(Filter):
    """Represents a valid string list constraint which should be handled as a containment filter"""

    field: Optional[List[str]]

    def to_query_type(self) -> Optional[Tuple[QueryType, object]]:
        if self.field:
            return (QueryType.CONTAINS, self.field)
        return None


class CombinedContainsFilterResourceState(Filter):
    """Represents a valid ReleasedResourceState constraint,
    which handles the filters as contains or not contains filters based on their values"""

    field: Optional[Dict[QueryType, List[ReleasedResourceState]]]

    @validator("field", pre=True)
    @classmethod
    def parse_field(cls, v: object) -> Optional[Dict[QueryType, ReleasedResourceState]]:
        if v is None:
            return None
        if isinstance(v, list) and all(isinstance(x, str) for x in v):
            status_contains_filters = [status_filter for status_filter in v if not status_filter.startswith("!")]
            status_not_contains_filters = [status_filter[1:] for status_filter in v if status_filter.startswith("!")]

            intersection = set(status_contains_filters).intersection(status_not_contains_filters)
            if len(intersection) > 0:
                raise ValueError(f"status expected to be both equal and not equal to: {intersection}")

            filters = {}
            if status_contains_filters:
                filters[QueryType.CONTAINS] = status_contains_filters
            if status_not_contains_filters:
                filters[QueryType.NOT_CONTAINS] = status_not_contains_filters

            return filters if filters else None

        raise ValueError(f"value is not a valid list of resource state constraints: {str(v)}")

    def to_query_type(self) -> Optional[Tuple[QueryType, object]]:
        if self.field:
            return (QueryType.COMBINED, self.field)
        return None


class ContainsFilterResourceAction(Filter):
    """Represents a valid ResourceAction list constraint which should be handled as a containment filter"""

    # Pydantic doesn't support Generic models on python 3.6
    field: Optional[List[const.ResourceAction]]

    def to_query_type(self) -> Optional[Tuple[QueryType, object]]:
        if self.field:
            return (QueryType.CONTAINS, self.field)
        return None


class LogLevelFilter(Filter):
    """Represents a valid LogLevel constraint which is considered to be the minimal log level"""

    field: Optional[const.LogLevel]

    @validator("field", pre=True)
    @classmethod
    def _field_single(cls, v: object) -> object:
        """
        Transform a list to a single log level
        """
        if isinstance(v, list):
            try:
                return const.LogLevel[
                    more_itertools.one(
                        v,
                        too_short=ValueError("Empty 'minimal_log_level' filter provided"),
                        too_long=ValueError(f"Multiple values provided for 'minimal_log_level' filter: {v}"),
                    ).upper()
                ]
            except KeyError:
                raise ValueError(f"{v} is not a valid log level")
        return v

    def to_query_type(self) -> Optional[Tuple[QueryType, object]]:
        if self.field is not None:
            return (QueryType.CONTAINS, self._get_log_levels_for_filter(self.field))
        return None

    def _get_log_levels_for_filter(self, minimal_log_level: const.LogLevel) -> List[str]:
        return [level.value for level in const.LogLevel if level.to_int >= minimal_log_level.to_int]


class FilterValidator(ABC):
    """
    This class provides methods to validate and process filters as received via the API.
    """

    @property
    @abstractmethod
    def allowed_filters(self) -> Dict[str, Type[Filter]]:
        """A dictionary that determines the mapping between the allowed filters and how they should be parsed and validated"""
        raise NotImplementedError()

    def process_filters(self, filter: Optional[Dict[str, List[str]]]) -> Dict[str, QueryFilter]:
        """
        Processes filters and returns a structured query filter object.

        :raises InvalidFilter: The supplied filter is invalid.
        """
        if filter is None:
            return {}
        query: Dict[str, QueryFilter] = {}
        for filter_name, filter_class in self.allowed_filters.items():
            try:
                # Validate the provided filter value with pydantic,
                # then determine how it should be handled according to its QueryType
                validated_query_type = filter_class(field=filter.get(filter_name)).to_query_type()
                if validated_query_type is not None:
                    query[filter_name] = validated_query_type
            except ValidationError as e:
                raise InvalidFilter(
                    f"Filter validation failed while parsing {filter_name}: {str(e)}, values provided: {str(filter)}"
                ) from e
        not_allowed_filters = set(filter.keys()) - set(self.allowed_filters.keys())
        if len(not_allowed_filters) > 0:
            raise InvalidFilter(f"The following filters are not supported: {not_allowed_filters}")
        return query
